#!/usr/bin/env python3
# -*- coding: utf-8 -*-


import copy
import dataclasses
import io
import pathlib
import shutil
import tarfile
import traceback
from typing import Optional, List, Tuple
import warnings

import numpy as np

import yaml

from ase import Atoms
from ase import units

from ase.io import read, write
import ase.constraints
from ase.constraints import Filter
from ase.optimize.optimize import Dynamics
from ase.md.velocitydistribution import (
    MaxwellBoltzmannDistribution,
    Stationary,
    ZeroRotation,
)

from ase.calculators.calculator import Calculator
from ase.calculators.singlepoint import SinglePointCalculator

from .. import config as GDPCONFIG
from ..potential.calculators.mixer import EnhancedCalculator
from .driver import AbstractDriver, DriverSetting, Controller
from .md.md_utils import force_temperature


def set_calc_state(calc: Calculator, timestep: float, stride: int):
    """Some calculators need driver information e.g. PLUMED."""
    if calc.name == "plumed":
        calc.timestep = timestep
        calc.stride = stride
    if hasattr(calc, "calcs"):
        for subcalc in calc.calcs:
            set_calc_state(subcalc, timestep, stride)
    else:
        ...

    return


def update_atoms_info(atoms: Atoms, dyn: Dynamics, start_step: int = 0) -> None:
    """Update step in atoms.info."""
    atoms.info["step"] = dyn.nsteps + start_step

    return


def retrieve_and_save_deviation(atoms, devi_fpath) -> None:
    """Read model deviation and add results to atoms.info if the file exists."""
    results = copy.deepcopy(atoms.calc.results)
    # devi_results = [(k,v) for k,v in results.items() if "devi" in k]
    devi_results = [
        (k, v) for k, v in results.items() if k in GDPCONFIG.VALID_DEVI_FRAME_KEYS
    ]
    if devi_results:
        devi_names = [x[0] for x in devi_results]
        devi_values = np.array([x[1] for x in devi_results]).reshape(1, -1)

        if devi_fpath.exists():
            with open(devi_fpath, "a") as fopen:
                np.savetxt(fopen, devi_values, fmt="%18.6e")
        else:
            with open(devi_fpath, "w") as fopen:
                np.savetxt(
                    fopen,
                    devi_values,
                    fmt="%18.6e",
                    header=("{:>18s}" * len(devi_names)).format(*devi_names),
                )

    return


def save_trajectory(atoms, log_fpath) -> None:
    """Create a clean atoms from the input and save simulation trajectory.

    We need an explicit copy of atoms as some calculators may not return all
    necessary information. For example, schnet only returns required properties.
    If only energy is required, there are no forces.

    """
    # - save atoms
    atoms_to_save = Atoms(
        symbols=atoms.get_chemical_symbols(),
        positions=atoms.get_positions().copy(),
        cell=atoms.get_cell().copy(),
        pbc=copy.deepcopy(atoms.get_pbc()),
    )
    if "tags" in atoms.arrays:
        atoms_to_save.set_tags(atoms.get_tags())
    if atoms.get_kinetic_energy() > 0.0:
        atoms_to_save.set_momenta(atoms.get_momenta())
    results = dict(
        energy=atoms.get_potential_energy(), forces=copy.deepcopy(atoms.get_forces())
    )
    spc = SinglePointCalculator(atoms, **results)
    atoms_to_save.calc = spc

    # - save atoms info...
    atoms_to_save.info["step"] = atoms.info["step"]

    # - save special keys and arrays from calc
    natoms = len(atoms)

    # -- add deviation
    for k, v in atoms.calc.results.items():
        if k in GDPCONFIG.VALID_DEVI_FRAME_KEYS:
            atoms_to_save.info[k] = v
    for k, v in atoms.calc.results.items():
        if k in GDPCONFIG.VALID_DEVI_ATOMIC_KEYS:
            atoms_to_save.arrays[k] = np.reshape(v, (natoms, -1))
    # print(f"keys: {atoms.calc.results.keys()}")

    # -- check special metadata
    calc = atoms.calc
    if isinstance(calc, EnhancedCalculator):
        atoms_to_save.info["host_energy"] = copy.deepcopy(calc.results["host_energy"])
        atoms_to_save.info["bias_energy"] = (
            results["energy"] - calc.results["host_energy"]
        )
        atoms_to_save.arrays["host_forces"] = copy.deepcopy(calc.results["host_forces"])

    # - append to traj
    write(log_fpath, atoms_to_save, append=True)

    return


def save_checkpoint(
    dyn: Dynamics, atoms: Atoms, wdir: pathlib.Path, ckpt_number: int=3,
    start_step: int = 0
):
    """"""
    ckpt_wdir = wdir / f"checkpoint.{dyn.nsteps+start_step}"
    ckpt_wdir.mkdir(parents=True, exist_ok=True)

    # write(ckpt_wdir/"structure.xyz", atoms)
    save_trajectory(atoms=atoms, log_fpath=ckpt_wdir / "structures.xyz")

    # For some optimisers and dynamics, they use random generator.
    if hasattr(dyn, "rng"):
        with open(ckpt_wdir / "rng_state.yaml", "w") as fopen:
            yaml.safe_dump(dyn.rng.bit_generator.state, fopen)

    # For some mixed calculator, save information, for example, PLUMED...
    if hasattr(atoms.calc, "calcs"):
        for calc in atoms.calc.calcs:
            if hasattr(calc, "_save_checkpoint"):
                calc._save_checkpoint(ckpt_wdir)

    # remove checkpoints if the number is over ckpt_number
    ckpt_wdirs = sorted(wdir.glob("checkpoint*"), key=lambda x: int(x.name[11:]))
    num_ckpts = len(ckpt_wdirs)
    if num_ckpts > ckpt_number:
        for w in ckpt_wdirs[:-ckpt_number]:
            shutil.rmtree(w)

    return


@dataclasses.dataclass
class BerendsenThermostat(Controller):

    name: str = "berendsen"

    def __post_init__(
        self,
    ):
        """"""
        taut = self.params.get("Tdamp", 100.0)
        assert taut is not None

        self.conv_params = dict(taut=taut * units.fs)

        return


@dataclasses.dataclass
class LangevinThermostat(Controller):

    name: str = "langevin"

    def __post_init__(
        self,
    ):
        """"""
        friction = self.params.get("friction", None)  # fs^-1
        assert friction is not None

        self.conv_params = dict(
            friction=friction / units.fs,
            # NOTE: The rng that generates friction normal distribution
            #       is set in `create_dynamics` by the driver's random_seed
            rng=None,
        )

        return


@dataclasses.dataclass
class NoseHooverThermostat(Controller):

    name: str = "nosehoover"

    def __post_init__(
        self,
    ):
        """"""
        nvt_q = self.params.get("nvt_q", None)
        assert nvt_q is not None

        self.conv_params = dict(nvt_q=nvt_q)  # thermostat mass

        return


@dataclasses.dataclass
class BerendsenBarostat(Controller):

    name: str = "berendsen"

    def __post_init__(
        self,
    ):
        """"""
        taut = self.params.get("Tdamp", 100.0)  # fs
        assert taut is not None

        taup = self.params.get("Pdamp", 100.0)  # fs
        assert taup is not None

        compressibility = self.params.get("compressibility", 4.57e-5)  # bar^-1

        self.conv_params = dict(
            taut=taut * units.fs,
            taup=taup * units.fs,
            compressibility_au=compressibility / units.bar,
        )

        return


@dataclasses.dataclass
class MonteCarloController(Controller):

    name: str = "monte_carlo"

    def __post_init__(self):
        """"""
        maxstepsize = self.params.get("maxstepsizes", 0.2)  # Ang

        self.conv_params = dict(
            maxstepsize=maxstepsize,
            rng=None,
        )

        return


controllers = dict(
    # - nvt
    berendsen_nvt=BerendsenThermostat,
    langevin_nvt=LangevinThermostat,
    nose_hoover_nvt=NoseHooverThermostat,
    monte_carlo_nvt=MonteCarloController,
    # - npt
    berendsen_npt=BerendsenBarostat,
)


@dataclasses.dataclass
class AseDriverSetting(DriverSetting):

    driver_cls: Optional[Dynamics] = None
    filter_cls: Optional[Filter] = None

    ensemble: str = "nve"

    controller: dict = dataclasses.field(default_factory=dict)

    fix_cm: bool = False

    fmax: float = 0.05  # eV/Ang

    def __post_init__(self):
        """"""
        # - task-specific params
        if self.task == "md":
            self._internals.update(
                # - helper params
                velocity_seed=self.velocity_seed,
                ignore_atoms_velocities=self.ignore_atoms_velocities,
            )
            if self.ensemble == "nve":
                from ase.md.verlet import VelocityVerlet as driver_cls

                _init_md_params = dict(
                    timestep=self.timestep * units.fs,
                )
            elif self.ensemble == "nvt":
                if self.controller:
                    thermo_cls_name = self.controller["name"] + "_" + self.ensemble
                    thermo_cls = controllers[thermo_cls_name]
                else:
                    thermo_cls = BerendsenThermostat
                thermostat = thermo_cls(**self.controller)
                if thermostat.name == "berendsen":
                    from ase.md.nvtberendsen import NVTBerendsen as driver_cls
                elif thermostat.name == "langevin":
                    from ase.md.langevin import Langevin as driver_cls
                elif thermostat.name == "nose_hoover":
                    from .md.nosehoover import NoseHoover as driver_cls
                elif thermostat.name == "monte_carlo":
                    from .mc.tfmc import TimeStampedMonteCarlo as driver_cls
                else:
                    raise RuntimeError(f"Unknown thermostat {thermostat}.")
                thermo_params = thermostat.conv_params
                _init_md_params = dict(
                    fixcm=self.fix_cm,
                    timestep=self.timestep * units.fs,
                    temperature_K=self.temp,
                )
                _init_md_params.update(**thermo_params)
            elif self.ensemble == "npt":
                if self.controller:
                    baro_cls_name = self.controller["name"] + "_" + self.ensemble
                    baro_cls = controllers[baro_cls_name]
                else:
                    baro_cls = BerendsenBarostat
                barostat = baro_cls(**self.controller)
                if barostat.name == "berendsen":
                    from ase.md.nptberendsen import NPTBerendsen as driver_cls
                else:
                    raise RuntimeError(f"Unknown barostat {barostat}.")
                baro_params = barostat.conv_params
                _init_md_params = dict(
                    fixcm=self.fix_cm,
                    timestep=self.timestep * units.fs,
                    temperature_K=self.temp,
                    pressure_au=self.press * (1e5 * units.Pascal),
                )
                _init_md_params.update(**baro_params)

            self._internals.update(**_init_md_params)

        if self.task == "min":
            # - to opt atomic positions
            from ase.optimize import BFGS

            if self.min_style == "bfgs":
                driver_cls = BFGS
            # - to opt unit cell
            #   UnitCellFilter, StrainFilter, ExpCellFilter
            # TODO: add filter params
            filter_names = ["unitCellFilter", "StrainFilter", "ExpCellFilter"]
            if self.min_style in filter_names:
                driver_cls = BFGS
                self.filter_cls = getattr(ase.constraints, self.min_style)

        if self.task == "rxn":
            # TODO: move to reactor
            try:
                from sella import Sella, Constraints

                driver_cls = Sella
            except:
                raise NotImplementedError(f"Sella is not installed.")
            ...

        try:
            self.driver_cls = driver_cls
        except:
            raise RuntimeError("Ase Driver Class is not defined.")

        # - shared params
        self._internals.update(loginterval=self.dump_period)

        # NOTE: There is a bug in ASE as it checks `if steps` then fails when spc.
        if self.steps == 0:
            self.steps = -1

        return

    def get_run_params(self, *args, **kwargs) -> dict:
        """"""
        run_params = dict(
            steps=kwargs.get("steps", self.steps),
            constraint=kwargs.get("constraint", self.constraint),
        )
        if self.task == "min" or self.task == "ts":
            run_params.update(
                fmax=kwargs.get("fmax", self.fmax),
            )
        run_params.update(**kwargs)

        return run_params


class AseDriver(AbstractDriver):

    name = "ase"

    # - defaults
    default_task = "min"
    supported_tasks = ["min", "rxn", "md"]

    # - other files
    log_fname = "dyn.log"
    xyz_fname = "traj.xyz"
    devi_fname = "model_devi-ase.dat"

    #: List of output files would be saved when restart.
    saved_fnames: List[str] = [log_fname, xyz_fname, devi_fname]

    #: List of output files would be removed when restart.
    removed_fnames: List[str] = [log_fname, xyz_fname, devi_fname]

    def __init__(self, calc=None, params: dict = {}, directory="./", *args, **kwargs):
        """"""
        super().__init__(calc, params, directory, *args, **kwargs)

        self.setting = AseDriverSetting(**params)

        self._log_fpath = self.directory / self.log_fname

        return

    @property
    def log_fpath(self):
        """File path of the simulation log."""

        return self._log_fpath

    @AbstractDriver.directory.setter
    def directory(self, directory_):
        """Set log and traj path regarding to the working directory."""
        # - main and calc
        super(AseDriver, AseDriver).directory.__set__(self, directory_)

        # - other files
        self._log_fpath = self.directory / self.log_fname

        return

    def _create_dynamics(self, atoms: Atoms, *args, **kwargs) -> Tuple[Dynamics, dict]:
        """Create the correct class of this simulation with running parameters.

        Respect `steps` and `fmax` as restart.

        """
        # - overwrite
        run_params = self.setting.get_run_params(*args, **kwargs)

        # -
        self._preprocess_constraints(atoms, run_params)

        # - init driver
        if self.setting.task == "min":
            if self.setting.filter_cls:
                atoms = self.setting.filter_cls(atoms)
            driver = self.setting.driver_cls(
                atoms, logfile=self.log_fpath, trajectory=None
            )
        elif self.setting.task == "ts":
            driver = self.setting.driver_cls(
                atoms, order=1, internal=False, logfile=self.log_fpath, trajectory=None
            )
        elif self.setting.task == "md":
            # - adjust params
            init_params_ = self.setting.get_init_params()

            # - velocity
            # NOTE: every dynamics will have a new rng...
            velocity_seed = init_params_.pop("velocity_seed")
            if velocity_seed is None:
                self._print(f"MD Driver's velocity_seed: {self.random_seed}")
                vrng = np.random.Generator(np.random.PCG64(self.random_seed))
            else:
                self._print(f"MD Driver's velocity_seed: {velocity_seed}")
                # vrng = np.random.default_rng(velocity_seed)
                vrng = np.random.Generator(np.random.PCG64(velocity_seed))

            ignore_atoms_velocities = init_params_.pop("ignore_atoms_velocities")
            if not ignore_atoms_velocities and atoms.get_kinetic_energy() > 0.0:
                # atoms have momenta
                ...
            else:
                MaxwellBoltzmannDistribution(
                    atoms, temperature_K=init_params_["temperature_K"], rng=vrng
                )
                if self.setting.remove_rotation:
                    ZeroRotation(atoms, preserve_temperature=False)
                if self.setting.remove_translation:
                    Stationary(atoms, preserve_temperature=False)
                # NOTE: respect constraints
                #       ase code does not consider constraints
                force_temperature(atoms, init_params_["temperature_K"], unit="K")

            # - some dynamics need rng
            if "rng" in init_params_:
                self._print(f"MD Driver's rng: {self.rng.bit_generator.state}")
                init_params_["rng"] = self.rng

            # - other callbacks
            set_calc_state(
                self.calc,
                timestep=init_params_["timestep"],
                stride=init_params_["loginterval"],
            )

            # - construct the driver
            driver = self.setting.driver_cls(
                atoms=atoms, **init_params_, logfile=self.log_fpath, trajectory=None
            )
        else:
            raise NotImplementedError(f"Unknown task {self.setting.task}.")

        return driver, run_params

    def _verify_checkpoint(self, *args, **kwargs) -> bool:
        """"""
        verified = super()._verify_checkpoint()
        if verified:
            asetraj = self.directory / self.xyz_fname
            if asetraj.exists() and asetraj.stat().st_size != 0:
                temp_frames = read(asetraj, ":")
                try:
                    _ = temp_frames[0].get_forces()
                except:  # `RuntimeError: Atoms object has no calculator.`
                    verified = False
            else:
                verified = False
        else:
            ...

        return verified

    def _find_latest_checkpoint(self, wdir: pathlib.Path):
        """"""
        ckpt_dirs = sorted(
            wdir.glob("checkpoint.*"), key=lambda x: int(x.name.split(".")[-1])
        )
        latest_ckpt_dir = ckpt_dirs[-1]

        return latest_ckpt_dir

    def _load_checkpoint(self, ckpt_dir: pathlib.Path):
        """"""
        atoms = read(ckpt_dir / "structures.xyz", ":")[-1]

        rng_state_fpath = ckpt_dir / "rng_state.yaml"
        if rng_state_fpath.exists():
            with open(ckpt_dir / "rng_state.yaml", "r") as fopen:
                rng_state = yaml.safe_load(fopen)
        else:
            rng_state = None

        return atoms, rng_state

    def _irun(
        self,
        atoms: Atoms,
        ckpt_wdir=None,
        cache_traj: List[Atoms] = None,
        *args,
        **kwargs,
    ):
        """Run the simulation."""
        prev_wdir = ckpt_wdir
        try:
            # To restart, velocities are always retained
            prev_ignore_atoms_velocities = self.setting.ignore_atoms_velocities
            if prev_wdir is None:  # start from the scratch
                start_step = 0
                rng_state = None

                curr_params = {}
                curr_params["random_seed"] = self.random_seed
                curr_params["init"] = self.setting.get_init_params()
                curr_params["run"] = self.setting.get_run_params()

                with open(self.directory / "params.yaml", "w") as fopen:
                    yaml.safe_dump(curr_params, fopen, indent=2)
            else:  # restart ...
                ckpt_wdir = self._find_latest_checkpoint(prev_wdir)
                atoms, rng_state = self._load_checkpoint(ckpt_wdir)
                start_step = atoms.info["step"]
                if hasattr(self.calc, "calcs"):
                    for calc in self.calc.calcs:
                        if hasattr(calc, "_load_checkpoint"):
                            calc._load_checkpoint(ckpt_wdir, start_step=start_step)
                # --- update run_params in settings
                target_steps = self.setting.get_run_params(*args, **kwargs)["steps"]
                if target_steps > 0:
                    steps = target_steps - start_step
                assert steps > 0, "Steps should be greater than 0."
                kwargs.update(steps=steps)

                # To restart, velocities are always retained
                self.setting.ignore_atoms_velocities = False

            # - set calculator
            atoms.calc = self.calc

            # - set dynamics
            dynamics, run_params = self._create_dynamics(atoms, *args, **kwargs)
            # dynamics.nsteps = start_step
            if hasattr(dynamics, "rng") and rng_state is not None:
                dynamics.rng.bit_generator.state = rng_state

            # --- callback functions
            dynamics.insert_observer(
                update_atoms_info,
                dyn=dynamics,
                atoms=atoms,
                start_step=start_step,
            )
            # NOTE: traj file not stores properties (energy, forces) properly
            init_params = self.setting.get_init_params()
            dynamics.attach(
                save_checkpoint,
                interval=self.setting.ckpt_period,
                dyn=dynamics,
                atoms=atoms,
                wdir=self.directory,
                ckpt_number=self.setting.ckpt_number,
                start_step=start_step,
            )
            dynamics.attach(
                save_trajectory,
                interval=init_params["loginterval"],
                atoms=atoms,
                log_fpath=self.directory / self.xyz_fname,
            )
            # NOTE: retrieve deviation info
            dynamics.attach(
                retrieve_and_save_deviation,
                interval=init_params["loginterval"],
                atoms=atoms,
                devi_fpath=self.directory / self.devi_fname,
            )
            dynamics.run(**run_params)

            # NOTE: check if the last frame is properly stored
            dump_period = self.setting.dump_period
            assert init_params["loginterval"] == dump_period

            if dump_period > 1:
                should_dump_last = False
                if self.setting.task == "min":
                    # optimiser dumps every step to log...
                    data = np.loadtxt(self.directory / "dyn.log", dtype=str, skiprows=1)
                    if len(data.shape) == 1:
                        data = data[np.newaxis, :]
                    nsteps = data.shape[0]
                    if nsteps > 0 and (nsteps - 1) % dump_period != 0:
                        should_dump_last = True
                elif self.setting.task == "md":
                    if dump_period > self.setting.steps:
                        should_dump_last = True
                if should_dump_last:
                    update_atoms_info(atoms, dynamics, start_step=start_step)
                    save_checkpoint(
                        dynamics, atoms, self.directory, ckpt_number=self.setting.ckpt_number,
                        start_step=start_step
                    )
                    save_trajectory(atoms, self.directory / self.xyz_fname)
                    retrieve_and_save_deviation(atoms, self.directory / self.devi_fname)
            else:
                ...

            # - Some interactive calculator needs kill processes after finishing,
            #   e.g. VaspInteractive...
            if hasattr(self.calc, "finalize"):
                self.calc.finalize()
            # To restart, velocities are always retained
            self.setting.ignore_atoms_velocities = prev_ignore_atoms_velocities
        except Exception as e:
            self._debug(f"Exception of {self.__class__.__name__} is {e}.")
            self._debug(
                f"Exception of {self.__class__.__name__} is {traceback.format_exc()}."
            )

        return

    def read_force_convergence(self, *args, **kwargs) -> bool:
        """Check if the force is converged.

        Sometimes DFT failed to converge SCF due to improper structure.

        """
        # - check convergence of forace evaluation (e.g. SCF convergence)
        scf_convergence = False
        try:
            scf_convergence = self.calc.read_convergence()
        except:
            # -- cannot read scf convergence then assume it is ok
            scf_convergence = True
        if not scf_convergence:
            warnings.warn(
                f"{self.name} at {self.directory} failed to converge at SCF.",
                RuntimeWarning,
            )
        # if not converged:
        #    warnings.warn(f"{self.name} at {self.directory} failed to converge.", RuntimeWarning)

        return scf_convergence

    def _read_a_single_trajectory(
        self, wdir: pathlib.Path, archive_path: pathlib.Path = None, *args, **kwargs
    ):
        """"""
        self._debug(f"archive_path: {archive_path}")
        self._debug(f"wdir: {wdir}")
        if archive_path is None:
            if (wdir / self.xyz_fname).exists():
                frames = read(wdir / self.xyz_fname, ":")
            else:
                frames = []
        else:
            target_name = str(
                (wdir / self.xyz_fname).relative_to(self.directory.parent)
            )
            with tarfile.open(archive_path, "r:gz") as tar:
                for tarinfo in tar:
                    if tarinfo.name == target_name:
                        fobj = io.StringIO(
                            tar.extractfile(tarinfo.name).read().decode()
                        )
                        frames = read(fobj, ":", format="extxyz")
                        fobj.close()
                        break
                else:
                    frames = []

        return frames

    def read_trajectory(self, archive_path=None, *args, **kwargs) -> List[Atoms]:
        """Read trajectory in the current working directory."""
        # - read trajectory
        traj_frames = self._aggregate_trajectories(archive_path=archive_path)

        # - add some info
        init_params = self.setting.get_init_params()
        if self.setting.task == "md":
            # Time[ps]      Etot[eV]     Epot[eV]     Ekin[eV]    T[K]
            # 0.0000           3.4237       2.8604       0.5633   272.4
            # data = np.loadtxt(self.directory/"dyn.log", dtype=float, skiprows=1)
            # if len(data.shape) == 1:
            #    data = data[np.newaxis,:]
            # timesteps = data[:, 0] # ps
            # steps = [int(s) for s in timesteps*1000/init_params["timestep"]]
            # ... infer from input settings
            for i, atoms in enumerate(traj_frames):
                atoms.info["time"] = (
                    i * init_params["timestep"] * init_params["loginterval"]
                )
        elif self.setting.task == "min":
            # Method - Step - Time - Energy - fmax
            # BFGS:    0 22:18:46    -1024.329999        3.3947
            # data = np.loadtxt(self.directory/"dyn.log", dtype=str, skiprows=1)
            # if len(data.shape) == 1:
            #    data = data[np.newaxis,:]
            # steps = [int(s) for s in data[:, 1]]
            # fmaxs = [float(fmax) for fmax in data[:, 4]]
            for atoms in traj_frames:
                atoms.info["fmax"] = np.max(
                    np.fabs(atoms.get_forces(apply_constraint=True))
                )
        # assert len(steps) == len(traj_frames), f"Number of steps {len(steps)} and number of frames {len(traj_frames)} are inconsistent..."

        # - deviation stored in traj, no need to read from file

        nframes = len(traj_frames)

        # calculation happens but some errors in calculation
        if self.directory.exists():
            # - check the convergence of the force evaluation
            try:
                scf_convergence = self.calc.read_convergence()
            except:
                # -- cannot read scf convergence then assume it is ok
                scf_convergence = True
            if not scf_convergence:
                warnings.warn(
                    f"{self.name} at {self.directory.name} failed to converge at SCF.",
                    RuntimeWarning,
                )
                if nframes > 0:  # for compat
                    traj_frames[0].info[
                        "error"
                    ] = f"Unconverged SCF at {self.directory}."
                traj_frames.error = True
        else:
            # TODO: How about archived data?
            ...

        return traj_frames


if __name__ == "__main__":
    ...



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import copy
import inspect
import time
import pathlib
from typing import Callable, List
import warnings

import numpy as np

import ase.formula

from ase import Atoms
from ase.io import read, write
from ase.ga.data import PrepareDB, DataConnection
from ase.ga.population import Population
from ase.ga.offspring_creator import OperationSelector

from .population import AbstractPopulationManager

from .. import registers
from .. import convert_indices
from ..expedition import AbstractExpedition

"""
TODO: search variational composition

Workflow
    check current calculation
        |
    check population and generate offspring
        |
    submit unrelaxed structures

Systems
    bulk
    slab
    cluster

Reserved Keywords in Database
    generation
    relaxed
    queued
    extinct
    description
    pairing

Operators
    comparator
    crossover (pairing)
    mutation

"""


class GeneticAlgorithemEngine(AbstractExpedition):
    """Genetic Algorithem Engine."""

    _directory = pathlib.Path.cwd()

    # local optimisation directory
    CALC_DIRNAME = "tmp_folder"

    #: Prefix of each generation's directory.
    GEN_PREFIX: str = "gen"

    # TODO: Neighbor list and parametrization parameters to screen
    # candidates before relaxation can be added. Default is not to use.
    find_neighbors = None
    perform_parametrization = None

    def __init__(
        self,
        builder: dict,
        params: dict,
        directroy="./",
        random_seed=None,
        *args,
        **kwargs,
    ):
        """Initialise engine.

        Args:
            builder: Define the system to explore.

        """
        ga_dict = params  # For compat

        # --- database ---
        self.db_name = ga_dict.get("database", "mydb.db")

        # -
        self.directory = directroy

        # -
        self.ga_dict = copy.deepcopy(ga_dict)

        # - random consistency, generator and population
        if random_seed is None:
            random_seed = np.random.randint(0, 1e8)
        self.random_seed = random_seed
        self._print(f"GA RANDOM SEED {random_seed}")
        self.rng = np.random.Generator(np.random.PCG64(seed=random_seed))

        # - check system type
        if isinstance(builder, dict):
            # -- generator will reset np.random by the given random_seed
            builder_params = copy.deepcopy(builder)
            builder_method = builder_params.pop("method")
            prev_seed = builder_params.get("random_seed")
            builder_params.update(random_seed=random_seed)
            self.generator = registers.create(
                "builder", builder_method, convert_name=False, **builder_params
            )
        else:
            prev_seed = builder.random_seed
            self.generator = builder
            np.random.seed(random_seed)
        self._print(f"OVERWRITE BUILDER SEED FROM {prev_seed} TO {random_seed}")

        # - worker info
        self.worker = None

        # --- population ---
        self.pop_manager = AbstractPopulationManager(
            ga_dict["population"], rng=self.rng
        )

        # --- property ---
        self.prop_dict = ga_dict["property"]
        target = self.prop_dict.get("target", "energy")
        self.prop_dict["target"] = target

        # --- convergence ---
        self.conv_dict = ga_dict["convergence"]

        # - misc
        self.use_archive = ga_dict.get("use_archive", True)

        return

    @property
    def directory(self):
        return self._directory

    @directory.setter
    def directory(self, directory_):
        self._directory = pathlib.Path(directory_)
        self.db_path = self._directory / self.db_name
        return

    def report(self):
        self._print("restart the database...")
        self.da = DataConnection(self.db_path)
        results = self.directory / "results"
        if not results.exists():
            results.mkdir()

        # - write structures
        all_relaxed_candidates = self.da.get_all_relaxed_candidates()
        write(results / "all_candidates.xyz", all_relaxed_candidates)

        # - plot population evolution
        data = []
        cur_gen_num = (
            self.da.get_generation_number()
        )  # equals finished generation plus one
        self._print(f"Current generation number: {cur_gen_num}")
        for i in range(cur_gen_num):
            # print('generation ', i)
            energies = [
                atoms.get_potential_energy()
                for atoms in all_relaxed_candidates
                if atoms.info["key_value_pairs"]["generation"] == i
            ]
            self._print(energies)
            data.append([i, energies])

        import matplotlib.pyplot as plt

        fig, ax = plt.subplots(nrows=1, ncols=1, figsize=(12, 8))
        ax.set_title("Population Evolution")

        for i, energies in data:
            ax.scatter([i] * len(energies), energies)

        fig.savefig(results / "pop.png")
        plt.close()

        return

    def update_active_params(self, prev_wdir, *args, **kwargs):
        """"""
        candidates = read(prev_wdir / "results" / "all_candidates.xyz", ":")

        init_size = self.pop_manager.init_size
        self.pop_manager.init_seed_file = candidates[:init_size]

        return

    def run(self, *args, **kwargs):
        """Run the GA procedure several steps.

        Default setting would run the algorithm many times until its convergence.
        This is useful for running optimisations with serial worker.

        """
        # - outputs
        assert self.worker is not None, "GA has not set its worker properly."
        self.worker.directory = self.directory / self.CALC_DIRNAME
        self.pop_manager._print = self._print

        # - search target
        self._print(f"===== Genetic Algorithm =====")
        target = self.prop_dict["target"]
        self._print(f"Target of Global Optimisation is {target}")

        # - worker info
        self._print("===== register worker =====")

        # - generator info
        self._print("===== register builder =====")
        for l in str(self.generator).split("\n"):
            self._print(l)
        self._print(f"random_state: f{self.generator.random_seed}")

        # NOTE: check database existence and generation number to determine restart
        self._debug(f"database path: {str(self.db_path)}")
        if not self.db_path.exists():
            self._print("----- create a new database -----")
            self._create_initial_population()
        else:
            self._print("restart the database...")
            self.da = DataConnection(self.db_path)

        # --- mutation and comparassion operators
        self._print("===== register operators =====")
        self._register_operators()

        # - run
        for i in range(1000):
            self._check_generation()
            if self.read_convergence():
                self._print("reach maximum generation...")
                self.report()
                break
            curr_convergence = self._irun()
            if not curr_convergence:
                self._print("current generation does not converge...")
                break

        return

    def _check_generation(self):
        """"""
        # --- check current generation number
        self.cur_gen = self.da.get_generation_number()
        # output a few info
        unrelaxed_strus_gen_ = list(
            self.da.c.select("relaxed=0,generation=%d" % self.cur_gen)
        )
        unrelaxed_strus_gen = []
        for row in unrelaxed_strus_gen_:
            # NOTE: mark_as_queue unrelaxed_candidate will have relaxed field too...
            if "queued" not in row:
                unrelaxed_strus_gen.append(row)
        self.unrelaxed_confids = [row["gaid"] for row in unrelaxed_strus_gen]
        self.num_unrelaxed_gen = len(self.unrelaxed_confids)

        relaxed_strus_gen = list(
            self.da.c.select("relaxed=1,generation=%d" % self.cur_gen)
        )
        for row in relaxed_strus_gen:
            self._debug(row)
        self.relaxed_confids = [row["gaid"] for row in relaxed_strus_gen]
        self.num_relaxed_gen = len(self.relaxed_confids)

        # check if this is the begin or the end of the current generation
        self.beg_of_gen = (self.num_relaxed_gen == self.num_unrelaxed_gen) and (
            self.num_relaxed_gen == 0
        )
        self.end_of_gen = (self.num_relaxed_gen == self.num_unrelaxed_gen) and (
            self.num_relaxed_gen != 0
        )

        return

    def _irun(self):
        """main procedure"""
        if not hasattr(self, "cur_gen"):
            raise RuntimeError(
                "The current genertion is unknown. Check generation before."
            )
        # - generation
        self._print("===== Generation Info =====")
        self._print(f"current generation number: {self.cur_gen}")
        self._print(f"number of relaxed in current generation: {self.num_relaxed_gen}")
        self._print(convert_indices(sorted(self.relaxed_confids)))
        self._print(
            f"number of unrelaxed in current generation: {self.num_unrelaxed_gen}"
        )
        self._print(convert_indices(sorted(self.unrelaxed_confids)))
        self._print(f"end of current generation: {self.end_of_gen}")

        # - population
        self._print("===== Population Info =====")
        content = "For generation > 0,\n"
        content += "{:>8s}  {:>8s}  {:>8s}  {:>8s}\n".format(
            "Reprod", "Random", "Mutate", "Total"
        )
        content += "{:>8d}  {:>8d}  {:>8d}  {:>8d}\n".format(
            self.pop_manager.gen_rep_size,
            self.pop_manager.gen_ran_size,
            self.pop_manager.gen_mut_size,
            self.pop_manager.gen_size,
        )
        content += "Note: Reproduced structure has a chance (pmut) to mutate.\n"
        for l in content.split("\n"):
            self._print(l)

        # - minimise
        if self.cur_gen == 0:
            self._print("===== Initial Population Calculation =====")
            frames_to_work = []
            while (
                self.da.get_number_of_unrelaxed_candidates()
            ):  # NOTE: this uses GADB get_atoms which adds extra_info
                # calculate structures from init population
                atoms = self.da.get_an_unrelaxed_candidate()
                frames_to_work.append(atoms)
                self.da.mark_as_queued(atoms)  # this marks relaxation is in the queue
            confids = [a.info["confid"] for a in frames_to_work]
            self._print(f"start to run structure {convert_indices(confids)}")
            # NOTE: provide unified interface to mlp and dft
            if frames_to_work:
                self.worker.directory = (
                    self.directory / self.CALC_DIRNAME / f"gen{self.cur_gen}"
                )
                _ = self.worker.run(frames_to_work)  # retrieve later
        else:
            # --- update population
            self._print("===== Update Population =====")
            # - create the population used for crossover and mutation
            candidate_groups, num_paired, num_mutated, num_random = (
                self.pop_manager._get_current_candidates(
                    database=self.da, curr_gen=self.cur_gen
                )
            )
            # candidate_groups, num_paired, num_mutated, num_random = self.pop_manager._get_current_candidates(
            #    database=self.da, curr_gen=self.cur_gen-1
            # )
            # for a in candidate_groups["paired"]:
            #    print(a.info)

            # TODO: random seed...
            current_population = Population(
                data_connection=self.da,
                population_size=self.pop_manager.gen_size,
                comparator=self.comparing,
                rng=np.random,  # This is set when the generator is created
            )
            self.pop_manager._update_generation_settings(
                current_population, self.mutations, self.pairing
            )

            # ----
            current_candidates = []
            if self.beg_of_gen:  # (num_relaxed == num_unrelaxed == 0)
                # -
                current_candidates = self.pop_manager._prepare_current_population(
                    database=self.da,
                    curr_gen=self.cur_gen,
                    population=current_population,
                    generator=self.generator,
                    pairing=self.pairing,
                    mutations=self.mutations,
                )
            else:
                self._print("Current generation has not finished...")
                # NOTE: The current candidates have not been created completely.
                #       For example, num_relaxed != num_unrelaxed,
                #       need create more candidates...
                if self.num_relaxed_gen == 0 and (
                    self.num_unrelaxed_gen < self.pop_manager.gen_size
                ):
                    current_candidates = self.pop_manager._prepare_current_population(
                        database=self.da,
                        curr_gen=self.cur_gen,
                        population=current_population,
                        generator=self.generator,
                        pairing=self.pairing,
                        mutations=self.mutations,
                        candidate_groups=candidate_groups,
                        num_paired=num_paired,
                        num_mutated=num_mutated,
                        num_random=num_random,
                    )
                elif self.num_relaxed_gen == 0 and (
                    self.num_unrelaxed_gen == self.pop_manager.gen_size
                ):
                    # no relaxed, and finished creation, num_relaxed == gen_size?
                    current_candidates = self.pop_manager._prepare_current_population(
                        database=self.da,
                        curr_gen=self.cur_gen,
                        population=current_population,
                        generator=self.generator,
                        pairing=self.pairing,
                        mutations=self.mutations,
                        candidate_groups=candidate_groups,
                        num_paired=num_paired,
                        num_mutated=num_mutated,
                        num_random=num_random,
                    )
                else:
                    ...

            # -- validate current candidates
            # candidate_groups, num_paired, num_mutated, num_random = self.pop_manager._get_current_candidates(
            #    database=self.da, curr_gen=self.cur_gen
            # )
            # for ia, a in enumerate(candidate_groups["paired"]):
            #    self._print(f"{ia} {a.info}")

            # TODO: send candidates directly to worker that respects the batchsize
            self._print("\n")
            self._print("\n")
            self._print("===== Optimisation =====")
            for ia, a in enumerate(current_candidates):
                self._print(f"{ia} {a.info}")
            if not (self.directory / self.CALC_DIRNAME / f"gen{self.cur_gen}").exists():
                frames_to_work = []
                for atoms in current_candidates:
                    frames_to_work.append(atoms)
                    self.da.mark_as_queued(
                        atoms
                    )  # this marks relaxation is in the queue
                if frames_to_work:
                    confids = [a.info["confid"] for a in frames_to_work]
                    self._print(f"start to run structure {convert_indices(confids)}")
                    self.worker.directory = (
                        self.directory / self.CALC_DIRNAME / f"gen{self.cur_gen}"
                    )
                    _ = self.worker.run(frames_to_work)  # retrieve later
            else:
                self._print(
                    f"calculation directory for generation {self.cur_gen} exists."
                )

        # --- check if there were finished jobs
        curr_convergence = False
        self.worker.directory = (
            self.directory / self.CALC_DIRNAME / f"gen{self.cur_gen}"
        )
        self.worker.inspect(resubmit=True)
        if self.worker.get_number_of_running_jobs() == 0:
            self._print("===== Retrieve Relaxed Population =====")
            converged_candidates = [
                t[-1] for t in self.worker.retrieve(use_archive=self.use_archive)
            ]
            for cand in converged_candidates:
                self._print(cand)
                self._print(cand.info)
                # update extra info
                extra_info = dict(data={}, key_value_pairs={"extinct": 0})
                cand.info.update(extra_info)
                # get tags
                confid = cand.info["confid"]
                if self.generator.use_tags:
                    rows = list(self.da.c.select(f"relaxed=0,gaid={confid}"))
                    for row in rows:
                        if row.formula:
                            previous_atoms = row.toatoms(
                                add_additional_information=True
                            )
                            previous_tags = previous_atoms.get_tags()
                            self._print(f"tags: {previous_tags}")
                            break
                    else:
                        raise RuntimeError(f"Cant find tags for cand {confid}")
                    cand.set_tags(previous_tags)
                # evaluate raw score
                self.evaluate_candidate(cand)
                self._print(f"  add relaxed cand {confid}")
                self._print(
                    "  with raw_score {:.4f}".format(
                        cand.info["key_value_pairs"]["raw_score"]
                    )
                )
                self.da.add_relaxed_step(
                    cand,
                    find_neighbors=self.find_neighbors,
                    perform_parametrization=self.perform_parametrization,
                )
            curr_convergence = True
        else:
            self._print("Worker is unfinished.")

        return curr_convergence

    def get_workers(self):
        """Get all workers used by this expedition."""
        if not hasattr(self, "da"):
            self.da = DataConnection(self.db_path)
            self._check_generation()

        num_gen = self.cur_gen
        if self.end_of_gen:
            num_gen += 1

        if hasattr(self.worker.potter, "remove_loaded_models"):
            self.worker.potter.remove_loaded_models()

        workers = []
        for i in range(num_gen):
            curr_worker = copy.deepcopy(self.worker)
            curr_worker.directory = (
                self.directory / self.CALC_DIRNAME / (f"{self.GEN_PREFIX}{i}")
            )
            workers.append(curr_worker)

        return workers

    def read_convergence(self):
        """check whether the search is converged"""
        if not hasattr(self, "cur_gen"):
            self.da = DataConnection(self.db_path)
            self._check_generation()
        max_gen = self.conv_dict["generation"]
        if self.cur_gen > max_gen and (self.num_relaxed_gen == self.num_unrelaxed_gen):
            return True
        else:
            return False

    def _create_operator(
        self, op_params: dict, specific_params: dict, mod_name: str, convert_name=False
    ):
        """Create operators such as comparator, crossover, and mutation.

        Args:
            op_params: Operator parameters loaded from input file.
            specific_params: Operator parameters obtained based on system.

        """
        op_params = copy.deepcopy(op_params)
        method = op_params.pop("method", None)
        if method is None:
            raise RuntimeError(f"There is no operator {method}.")
        op_cls = registers.get(mod_name, method, convert_name=convert_name)
        init_args = inspect.getargspec(op_cls.__init__).args[1:]  # skip self
        for k, v in specific_params.items():
            if k in init_args:
                op_params.update(**{k: v})
        op = op_cls(**op_params)

        return op

    def _register_operators(self):
        """register various operators
        comparator, pairing, mutations
        """
        op_dict = copy.deepcopy(self.ga_dict.get("operators", None))
        if op_dict is None:
            op_dict = {
                "comparator": {"name": "InteratomicDistanceComparator"},
                "crossover": {"name": "CutAndSplicePairing"},
            }

        specific_params = dict(
            slab=self.da.get_slab(),
            n_top=len(self.da.get_atom_numbers_to_optimize()),
            blmin=self.generator.blmin,
            number_of_variable_cell_vectors=self.generator.number_of_variable_cell_vectors,
            cell_bounds=self.generator.cell_bounds,
            test_dist_to_slab=self.generator.test_dist_to_slab,
            use_tags=self.generator.use_tags,
            used_modes_file=self.directory / self.CALC_DIRNAME / "used_modes.json",
            # rng = self.rng # TODO: ase operators need np.random
        )

        # --- comparator
        comp_params = op_dict.get("comparator", None)
        self.comparing = self._create_operator(
            comp_params, specific_params, "comparator", convert_name=True
        )

        self._print("--- comparator ---")
        self._print(f"Use comparator {self.comparing.__class__.__name__}.")

        # --- crossover
        crossover_params = op_dict.get("crossover", None)
        self.pairing = self._create_operator(
            crossover_params, specific_params, "builder", convert_name=False
        )

        self._print("--- crossover ---")
        self._print(f"Use crossover {self.pairing.__class__.__name__}.")
        # self._print("pairing: ", self.pairing)

        # --- mutations
        mutations, probs = [], []
        mutation_list = op_dict.get("mutation", [])
        if not isinstance(mutation_list, list):
            mutation_list = [mutation_list]
        for mut_params in mutation_list:
            prob = mut_params.pop("prob", 1.0)
            probs.append(prob)
            mut = self._create_operator(
                mut_params, specific_params, "builder", convert_name=False
            )
            mutations.append(mut)

        self._print("--- mutations ---")
        # self._print(f"mutation probability: {self.pmut}")
        for mut, prob in zip(mutations, probs):
            self._print(f"Use mutation {mut.descriptor} with prob {prob}.")
        self.mutations = OperationSelector(probs, mutations, rng=np.random)

        return

    def _create_initial_population(
        self,
    ):
        # create the database to store information in
        # TODO: move this part to where before generator is created
        da = PrepareDB(
            db_file_name=self.db_path,
            simulation_cell=self.generator._substrate,
            stoichiometry=self.generator.composition_atom_numbers,
        )

        starting_population = self.pop_manager._prepare_initial_population(
            generator=self.generator
        )

        self._print(f"save population {len(starting_population)} to database")
        for a in starting_population:
            da.add_unrelaxed_candidate(a)

        # TODO: change this to the DB interface
        self._print(
            "save population size {0} into database...".format(
                self.pop_manager.gen_size
            )
        )
        row = da.c.get(1)
        new_data = row["data"].copy()
        new_data["population_size"] = self.pop_manager.gen_size
        da.c.update(1, data=new_data)

        self.da = DataConnection(self.db_path)

        return

    def evaluate_candidate(self, atoms):
        """TODO: evaluate candidate based on raw score
        in most cases, it's potential energy
        but this is should be more flexible
        e.g. enthalpy (pot+pressure), reaction energy
        """
        assert (
            atoms.info["key_value_pairs"].get("raw_score", None) is None
        ), "candidate already has raw_score before evaluation"

        # NOTE: larger raw_score, better candidate

        # evaluate based on target property
        target = self.prop_dict["target"]
        if target == "energy":
            energy = atoms.get_potential_energy()
            forces = atoms.get_forces()
            atoms.info["key_value_pairs"]["raw_score"] = -energy
            from ase.build import niggli_reduce
            from ase.calculators.singlepoint import SinglePointCalculator

            if self.generator.cell_bounds:
                stress = atoms.get_stress()
                niggli_reduce(atoms)
                calc = SinglePointCalculator(
                    atoms, energy=energy, forces=forces, stress=stress
                )
                atoms.calc = calc
                if self.generator.cell_bounds.is_within_bounds(atoms.get_cell()):
                    atoms.info["key_value_pairs"]["raw_score"] = -energy
                else:
                    atoms.info["key_value_pairs"]["raw_score"] = -1e8
        elif target == "barrier":
            pass
        else:
            raise RuntimeError(f"Unknown target {target}...")

        return

    def as_dict(self) -> dict:
        """"""
        engine_params = {}
        engine_params["random_seed"] = self.random_seed
        engine_params["method"] = "genetic_algorithm"
        engine_params["builder"] = self.generator.as_dict()
        engine_params["worker"] = self.worker.as_dict()
        engine_params["params"] = self.ga_dict

        engine_params = copy.deepcopy(engine_params)

        return engine_params


if __name__ == "__main__":
    ...



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


import copy
import itertools
import pathlib

from typing import Optional, List, Tuple

import numpy as np

from ase import Atoms
from ase.data import atomic_numbers, covalent_radii
from ase.calculators.calculator import Calculator
from ase.neighborlist import NeighborList, natural_cutoffs

from .utils import get_bond_information


def compute_boost_energy_and_forces(
    positions,
    distances,
    shifts,
    ref_distances,
    bond_pairs,
    bond_strains,
    max_index: int,
    vmax=0.5,
    smax=0.5,
    curv: float = 0.98,
):
    """Compute bond-boost energy.

    u  = (1-(eps_max/smax)^2)
    dU = -2*(eps_max/smax^2)
    v  = (1-curv^2*(eps_max/smax)^2)
    dV = -2*curv^2*eps_max/smax^2

    The envelope function:
        A = u * u / v

    Args:
        ...

    Examples:
        >>>

    """
    num_bonds = len(bond_pairs)
    num_atoms = positions.shape[0]

    # - compute energy
    # V_b, shape (num_bonds, )
    vboost = vmax / num_bonds * (1 - (bond_strains / smax) ** 2)

    max_strain_ratio = bond_strains[max_index] / smax
    u = 1 - max_strain_ratio**2
    v = 1 - curv**2 * max_strain_ratio**2

    env = u * u / v
    energy = np.sum(vboost * env)

    # - compute forces
    forces = np.zeros((num_atoms, 3))

    # dV_b/deps_i, shape (num_bonds, )
    d_vboost = -vmax / num_bonds * 2 * bond_strains / smax**2

    # -- shared terms
    for p, (i, j) in enumerate(bond_pairs):
        frc_ij = (
            -env
            * d_vboost[p]
            * (positions[i] - (positions[j] + shifts[p]))
            / distances[p]
            / ref_distances[p]
        )
        forces[i] += frc_ij
        forces[j] += -frc_ij

    # -- the extra term for max_index
    du = -2 * max_strain_ratio / smax
    dv = -2 * curv**2 * max_strain_ratio / smax

    denv = du * (u / v) + u * (du * v - u * dv) / v**2

    max_i, max_j = bond_pairs[max_index]
    max_frc_ij = (
        -denv
        * (positions[max_i] - positions[max_j])
        / distances[max_index]
        / ref_distances[max_index]
        * np.sum(vboost)
    )
    forces[max_i] += max_frc_ij
    forces[max_j] += -max_frc_ij

    return energy, forces


class BondBoostCalculator(Calculator):

    implemented_properties = ["energy", "free_energy", "forces"]

    def __init__(
        self,
        curv: float = 0.98,
        vmax: float = 0.5,
        smax: float = 0.5,
        bonds: List[str] = ["C", "H", "O"],
        cov_equi_ratio: float = 1.0,
        covalent_ratio: Tuple[float, float] = [0.8, 1.6],
        target_indices: Optional[List[int]] = None,
        *args,
        **kwargs,
    ):
        """"""
        super().__init__(*args, **kwargs)

        # bond-boost params
        #:V_max, eV
        self.vmax = vmax

        #: control the curvature near the boundary
        self.curv = curv

        #: q, maximum bond change compared to the reference state
        self.smax = smax

        self.covalent_ratio = covalent_ratio
        if not np.isclose(self.covalent_ratio[1], 1.0 + self.smax):
            self.covalent_ratio[1] = 1.0 + self.smax

        # aux params
        #: NeighborList
        self.neighlist = None

        #: Equilibrium bond distance dict.
        if isinstance(bonds[0], str):
            symbols = bonds
            bonds = list(itertools.product(bonds, bonds))
        else:
            symbols_, bonds_ = [], []
            for i, j in bonds:
                symbols_.extend([i, j])
                bonds_.append((i, j))
                bonds_.append((j, i))
            symbols = list(set(symbols_))
            bonds = list(set(bonds_))
        self.symbols = symbols
        self.bonds = bonds

        radii = {s: cov_equi_ratio*covalent_radii[atomic_numbers[s]] for s in symbols}
        self.eqdis_dict = {k: radii[k[0]] + radii[k[1]] for k in bonds}

        #:
        if target_indices is not None:
            self.target_indices = target_indices
        else:
            self.target_indices = None

        return

    def calculate(
        self,
        atoms: Optional[Atoms] = None,
        properties=["energy"],
        system_changes=["positions"],
    ):
        """"""
        super().calculate(atoms, properties, system_changes)

        log_fpath = pathlib.Path(self.directory) / "info.log"
        if not log_fpath.exists():
            content = f"# {self.vmax =} {self.smax =} {self.curv = }\n"
            content += f"# {self.covalent_ratio[0]} {self.covalent_ratio[1]}\n"
            content += f"# {self.eqdis_dict}"
            content += f"# num_bonds max_i max_j max_strain\n"
            with open(log_fpath, "w") as fopen:
                fopen.write(content)

        if self.neighlist is None:
            covalent_max = self.covalent_ratio[
                1
            ]  # (1+smax) times covalent bond distance
            self.neighlist = NeighborList(
                covalent_max * np.array(natural_cutoffs(atoms)),
                skin=0.0,
                self_interaction=False,
                bothways=False,
            )
        else:
            ...
        self.neighlist.update(atoms)

        # - find species
        if self.target_indices is None:
            self.target_indices = [i for i, a in enumerate(atoms) if a.symbol in self.symbols]

        # - find bonds to boost
        bond_pairs, bond_distances, bond_shifts, equi_distances = get_bond_information(
            atoms,
            self.neighlist,
            self.eqdis_dict,
            covalent_min=self.covalent_ratio[0],
            target_indices=self.target_indices,
            allowed_bonds=self.bonds,
        )

        # - compute properties
        num_bonds = len(bond_pairs)
        if num_bonds > 0:
            # - compute properties
            bond_strains = (bond_distances - equi_distances) / equi_distances
            max_index = np.argmax(bond_strains)

            energy, forces = compute_boost_energy_and_forces(
                atoms.positions,
                bond_distances,
                bond_shifts,
                equi_distances,
                bond_pairs,
                bond_strains,
                max_index,
                vmax=self.vmax,
                smax=self.smax,
                curv=self.curv,
            )
            # print(f"{max_index =}")
            # print(f"{bond_pairs[max_index] =}")
            # print(f"{bond_distances[max_index] =}")
            # print(f"{bond_strains[max_index] =}")
            self._write_step(
                num_bonds,
                [atoms[x].symbol for x in bond_pairs[max_index]],
                bond_pairs[max_index],
                bond_distances[max_index],
                bond_strains[max_index],
            )
        else:
            energy = 0.0
            forces = np.zeros((atoms.positions.shape))
            self._write_step(num_bonds, [np.nan, np.nan], [np.nan, np.nan], np.nan, np.nan)

        self.results["energy"] = energy
        self.results["free_energy"] = energy
        self.results["forces"] = forces

        return

    def _write_step(
        self,
        num_bonds: int,
        bond_symb: Tuple[str, str],
        bond_pair: Tuple[int, int],
        distance: float,
        strain: float,
    ):
        """"""
        pair = "-".join([str(s) + "_" + str(x) for s, x in zip(bond_symb, bond_pair)])
        content = f"{num_bonds:>12d}  {pair:>24s}  {distance:>12.4f}  {strain:>12.4f}\n"

        log_fpath = pathlib.Path(self.directory) / "info.log"
        with open(log_fpath, "a") as fopen:
            fopen.write(content)

        return


if __name__ == "__main__":
    ...



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import pathlib
from typing import List

import numpy as np

import matplotlib
matplotlib.use("Agg") #silent mode
import matplotlib.pyplot as plt
try:
    plt.style.use("presentation")
except Exception as e:
    #print("Used default matplotlib style.")
    ...

from ase import Atoms
from ase.io import read, write
from ase.eos import EquationOfState 

from .validator import AbstractValidator
from ..data.array import AtomsNDArray

"""Validation on equation of states.
"""

def plot_eos(ref_volumes, ref_energies, pre_volumes, pre_energies, fig_fpath="./bm.png"):
    """"""
    # plot figure
    fig, ax = plt.subplots(nrows=1, ncols=1, figsize=(12,8))
    ax.set_title("Birch-Murnaghan (Constant-Volume Optimisation)")
    
    ax.set_xlabel("Volume/atom [Å^3/atom]")
    ax.set_ylabel("Energy/atom [eV/atom]")

    ax.scatter(ref_volumes, ref_energies, marker="*", label="reference")
    ax.scatter(pre_volumes, pre_energies, marker="*", label="prediction")

    ax.legend()

    plt.savefig(fig_fpath)

    return


class EquationOfStateValidator(AbstractValidator):

    def __init__(self, *args, **kwargs):
        """"""
        super().__init__(*args, **kwargs)

        return
    
    def _process_data(self, data) -> List[Atoms]:
        """"""
        data = AtomsNDArray(data)

        if data.ndim == 1:
            data = data.tolist()
        elif data.ndim == 2: # assume it is from minimisations...
            data = data[:, -1]
        else:
            raise RuntimeError(f"Invalid shape {data.shape}.")

        return data
    
    def run(self, dataset, worker=None, *args, **kwargs):
        """"""
        # - preprocess the dataset
        pre_dataset = dataset.get("prediction", None) # Assume it is ndarray
        pre_dataset = self._process_data(pre_dataset)

        ref_dataset = dataset.get("reference", None)
        ref_dataset = self._process_data(ref_dataset)
        
        # - 
        ref_natoms = np.array([len(a) for a in ref_dataset])
        ref_volumes = np.array([a.get_volume() for a in ref_dataset]) / ref_natoms
        ref_energies = np.array([a.get_potential_energy() for a in ref_dataset]) / ref_natoms

        pre_natoms = np.array([len(a) for a in pre_dataset])
        pre_volumes = np.array([a.get_volume() for a in pre_dataset]) / pre_natoms
        pre_energies = np.array([a.get_potential_energy() for a in pre_dataset]) / pre_natoms

        plot_eos(
            ref_volumes, ref_energies, pre_volumes, pre_energies, 
            fig_fpath=self.directory/"bm.png"
        )

        return


if __name__ == "__main__":
    ...


#!/usr/bin/env python3
# -*- coding: utf-8 -*-


from .deepmd import DeepmdManager, DeepmdDataloader, DeepmdTrainer

__all__ = [
    "DeepmdManager", "DeepmdDataloader", "DeepmdTrainer"
]


if __name__ == "__main__":
    ...



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import copy
import itertools
from typing import List
import warnings

import numpy as np

import matplotlib as mpl
mpl.use("Agg") #silent mode
from matplotlib import pyplot as plt
try:
    plt.style.use("presentation")
except Exception as e:
    #print("Used default matplotlib style.")
    ...

from ase import Atoms
from ase.io import read, write

from ..worker.drive import DriverBasedWorker
from .validator import AbstractValidator
from .utils import get_properties


class DimerValidator(AbstractValidator):

    def run(self, dataset, worker: DriverBasedWorker, *args, **kwargs):
        """"""
        super().run()
        self._print(dataset["reference"])
        for prefix, frames in dataset["reference"].items():
            self._irun(prefix, frames, None, worker)

        return

    def _irun(self, prefix: str, ref_frames: List[Atoms], pred_frames: List[Atoms], worker):
        """"""
        # - check if input frames are dimers...
        chemicals = []
        for atoms in ref_frames:
            natoms = len(atoms)
            assert natoms == 2, f"Input structure at {self.directory} must be a dimer."
            chemicals.append(atoms.get_chemical_formula())
            assert len(set(chemicals)) == 1, f"Input dimers are different. Maybe {chemicals[0]}?"
        
        # - read reference
        ref_symbols, ref_energies, ref_forces = get_properties(ref_frames)
        nframes = len(ref_frames)
        ref_natoms = [len(a) for a in ref_frames]

        if pred_frames is None:
            # NOTE: use worker to calculate
            self._print(f"Calculate reference frames {prefix} with potential...")
            cached_pred_fpath = self.directory / prefix / "pred.xyz"
            if not cached_pred_fpath.exists():
                worker.directory = self.directory / prefix
                worker.batchsize = nframes

                worker._share_wdir = True

                worker.run(ref_frames)
                worker.inspect(resubmit=True)
                if worker.get_number_of_running_jobs() == 0:
                    ret = worker.retrieve(
                        include_retrieved=True,
                    )
                    pred_frames = list(itertools.chain(*ret))
                else:
                    ...
                write(cached_pred_fpath, pred_frames)
            else:
                pred_frames = read(cached_pred_fpath, ":")
        else:
            ...
        pred_symbols, pred_energies, pred_forces = get_properties(pred_frames)
    
        # - get reaction coordinate (dimer distance here)
        ref_distances = []
        for atoms in ref_frames:
            dis = atoms.get_distance(0, 1, mic=True)
            ref_distances.append(dis)
        
        # - save data
        abs_errors = [x-y for x,y in zip(pred_energies, ref_energies)]
        rel_errors = [(x/y)*100. for x,y in zip(abs_errors,ref_energies)]
        data = np.array([ref_distances,ref_energies,pred_energies,abs_errors,rel_errors]).T

        np.savetxt(
            self.directory/prefix/f"{prefix}.dat", data, 
            fmt="%8.4f  %12.4f  %12.4f  %12.4f  %8.4f", 
            header="{:<8s}  {:<12s}  {:<12s}  {:<12s}  {:<8s}".format(
                "dis", "ref", "mlp", "abs", "rel [%]"
            )
        )

        # - plot data
        fig, ax = plt.subplots(
            nrows=1, ncols=1,
            gridspec_kw={"hspace": 0.3}, figsize=(16, 9)
        )
        plt.suptitle(f"{prefix} with nframes {nframes}")

        ax.plot(ref_distances, ref_energies, marker="o",  markerfacecolor="w", label="Reference")
        ax.plot(ref_distances, pred_energies, marker="o", markerfacecolor="w", label="Prediction")

        ax.set_ylabel("Energy [eV]")
        ax.set_xlabel("Dimer Distance [Å]")

        ax.legend()

        if (self.directory/f"{prefix}.png").exists():
            warnings.warn(f"Figure file {prefix} exists.", UserWarning)
        else:
            plt.savefig(self.directory/prefix/f"{prefix}.png")

        return
    

if __name__ == "__main__":
    pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import os
import copy
import io
import itertools
import shutil
import warnings
import subprocess
import pathlib
import pickle
import tarfile
import dataclasses

from collections.abc import Iterable
from typing import List, Mapping, Dict, Optional, NoReturn, Tuple

import numpy as np

from ase import Atoms
from ase import units
from ase.data import atomic_numbers, atomic_masses
from ase.io import read, write
from ase.io.lammpsrun import read_lammps_dump_text
from ase.io.lammpsdata import read_lammps_data, write_lammps_data
from ase.calculators.lammps import unitconvert, Prism
from ase.calculators.calculator import (
    CalculationFailed,
    Calculator, all_changes, PropertyNotImplementedError, FileIOCalculator
)
from ase.calculators.singlepoint import SinglePointCalculator
from ase.calculators.mixing import LinearCombinationCalculator

from .. import config
from ..builder.constraints import parse_constraint_info
from .driver import AbstractDriver, DriverSetting

from ..potential.managers.plumed.calculators.plumed2 import Plumed, update_stride_and_file


dataclasses.dataclass(frozen=True)
class AseLammpsSettings:

    """File names."""

    inputstructure_filename: str = "stru.data"
    trajectory_filename: str = "traj.dump"
    input_fname: str = "in.lammps"
    log_filename: str = "log.lammps"
    deviation_filename: str = "model_devi.out"
    prism_filename: str = "ase-prism.bindat"

#: Instance.
ASELMPCONFIG = AseLammpsSettings()

def parse_type_list(atoms):
    """Parse the type list based on input atoms."""
    # elements
    type_list = list(set(atoms.get_chemical_symbols()))
    type_list.sort() # by alphabet

    return type_list

def parse_thermo_data(lines) -> dict:
    """Read energy ... results from log.lammps file."""
    # - parse input lines
    found_error = False
    start_idx, end_idx = None, None
    for idx, line in enumerate(lines):
        # - get the line index at the start of the thermo infomation
        #   test with 29Oct2020 and 23Jun2022
        if line.strip().startswith("Step"):
            start_idx = idx
        # - NOTE: find line index at the end
        if line.strip().startswith("ERROR: "):
            found_error = True
            end_idx = idx
        if line.strip().startswith("Loop time"):
            end_idx = idx
        if start_idx is not None and end_idx is not None:
            break
    else:
        end_idx = idx
    config._debug(f"Initial lammps LOG index: {start_idx} {end_idx}")
    
    # - check valid lines
    #   sometimes the line may not be complete
    ncols = len(lines[start_idx].strip().split())
    for i in range(end_idx, start_idx, -1):
        curr_data = lines[i].strip().split()
        curr_ncols = len(curr_data)
        config._debug(f"Error: {lines[i]}")
        if curr_ncols == ncols: # still log step info and no LOOP
            try:
                step = int(curr_data[0])
                end_idx = i+1
            except ValueError:
                config._debug(f"Error: {lines[i]}")
            finally:
                break
        else:
            ...
    else:
        end_idx = None # even not one single complete line
    config._debug(f"Sanitised lammps LOG index: {start_idx} {end_idx}")

    if start_idx is None or end_idx is None:
        raise RuntimeError(f"Error in lammps output with start {start_idx} end {end_idx}.")
    end_info = lines[end_idx] # either loop time or error
    config._debug(f"lammps END info: {end_info}")

    config._debug(f"lammps LOG index: {start_idx} {end_idx}")

    # -- parse index of PotEng
    # TODO: save timestep info?
    thermo_keywords = lines[start_idx].strip().split()
    if "PotEng" not in thermo_keywords:
        raise RuntimeError(f"Cant find PotEng in lammps output.")
    thermo_data = []
    for x in lines[start_idx+1:end_idx]:
        x_data = x.strip().split()
        if x_data[0].isdigit(): # There may have some extra warnings... such as restart
            thermo_data.append(x_data)
    #thermo_data = np.array([line.strip().split() for line in thermo_data], dtype=float).transpose()
    thermo_data = np.array(thermo_data, dtype=float).transpose() 
    #config._debug(thermo_data)
    thermo_dict = {}
    for i, k in enumerate(thermo_keywords):
        thermo_dict[k] = thermo_data[i]

    return thermo_dict, end_info

def read_single_simulation(
        mdir, wdir: pathlib.Path, prefix: str, units: str, add_step_info=True,
        archive_path: pathlib.Path=None
    ):
    """"""
    # - get FileIO
    if archive_path is None:
        traj_io = open(wdir/ASELMPCONFIG.trajectory_filename, "r")
        log_io = open(wdir/ASELMPCONFIG.log_filename, "r")
        prism_file = wdir/ASELMPCONFIG.prism_filename
        if prism_file.exists():
            prism_io = open(prism_file, "rb")
        else:
            prism_io = None
        devi_path = wdir / (prefix+ASELMPCONFIG.deviation_filename)
        if devi_path.exists():
            devi_io = open(devi_path, "r")
        else:
            devi_io = None
        colvar_path = wdir/"COLVAR"
        if colvar_path.exists():
            colvar_io = open(colvar_path, "r")
        else:
            colvar_io = None
    else:
        rpath = wdir.relative_to(mdir.parent)
        traj_tarname = str(rpath/ASELMPCONFIG.trajectory_filename)
        prism_tarname = str(rpath/ASELMPCONFIG.prism_filename)
        log_tarname = str(rpath/ASELMPCONFIG.log_filename)
        devi_tarname = str(rpath/ASELMPCONFIG.deviation_filename)
        colvar_tarname = str(rpath/"COLVAR")
        prism_io, devi_io, colvar_io = None, None, None
        with tarfile.open(archive_path, "r:gz") as tar:
            for tarinfo in tar:
                if tarinfo.name.startswith(wdir.name):
                    if tarinfo.name == traj_tarname:
                        traj_io = io.StringIO(tar.extractfile(tarinfo.name).read().decode())
                    elif tarinfo.name == prism_tarname:
                        prism_io = io.BytesIO(tar.extractfile(tarinfo.name).read())
                    elif tarinfo.name == log_tarname:
                        log_io = io.StringIO(tar.extractfile(tarinfo.name).read().decode())
                    elif tarinfo.name == devi_tarname:
                        devi_io = io.StringIO(tar.extractfile(tarinfo.name).read().decode())
                    elif tarinfo.name == colvar_tarname:
                        colvar_io = io.StringIO(tar.extractfile(tarinfo.name).read().decode())
                    else:
                        ...
                else:
                    continue
            else: # TODO: if not find target traj?
                ...

    # - read timesteps
    timesteps = []
    while True:
        line = traj_io.readline()
        if "TIMESTEP" in line:
            timesteps.append(int(traj_io.readline().strip()))
        if not line:
            break
    traj_io.seek(0)

    # - read structure trajectory
    if prism_io is not None:
        prismobj = pickle.load(prism_io)
    else:
        prismobj = None

    curr_traj_frames_ = read(
        traj_io, 
        index=":", format="lammps-dump-text", prismobj=prismobj, units=units
    )
    nframes_traj = len(curr_traj_frames_)
    timesteps = timesteps[:nframes_traj] # avoid incomplete structure

    # - read thermo data
    thermo_dict, end_info = parse_thermo_data(log_io.readlines())

    # NOTE: last frame would not be dumpped if timestep not equals multiple*dump_period
    #       if there were any error, 
    pot_energies = [unitconvert.convert(p, "energy", units, "ASE") for p in thermo_dict["PotEng"]]
    nframes_thermo = len(pot_energies)
    nframes = min([nframes_traj, nframes_thermo])
    config._debug(f"nframes in lammps: {nframes} traj {nframes_traj} thermo {nframes_thermo}")

    # NOTE: check whether steps in thermo and traj are consistent
    #pot_energies = pot_energies[:nframes]
    #curr_traj_frames = curr_traj_frames[:nframes]
    #assert len(pot_energies) == len(curr_traj_frames), f"Number of pot energies and frames are inconsistent at {str(wdir)}."

    curr_traj_frames, curr_energies = [], []
    for i, t in enumerate(timesteps):
        if t in thermo_dict["Step"]:
            curr_atoms = curr_traj_frames_[i]
            curr_atoms.info["step"] = t
            curr_traj_frames.append(curr_atoms)
            curr_energies.append(pot_energies[thermo_dict["Step"].tolist().index(t)])

    for pot_eng, atoms in zip(curr_energies, curr_traj_frames):
        forces = atoms.get_forces()
        # NOTE: forces have already been converted in ase read, so velocities are
        sp_calc = SinglePointCalculator(atoms, energy=pot_eng, forces=forces)
        atoms.calc = sp_calc

    # - check model_devi.out
    # TODO: convert units?
    if devi_io is not None:
        lines = devi_io.readlines()
        if "#" in lines[0]: # the first file
            dkeys = ("".join([x for x in lines[0] if x != "#"])).strip().split()
            dkeys = [x.strip() for x in dkeys][1:]
        else:
            ...
        devi_io.seek(0)
        data = np.loadtxt(devi_io, dtype=float)
        ncols = data.shape[-1]
        data = data.reshape(-1, ncols)
        # NOTE: For some minimisers, dp gives several deviations as 
        #       multiple force evluations are performed in one step.
        #       Thus, we only take the last occurance of the deviation in each step.
        step_indices = []
        steps = data[:, 0].astype(np.int32).tolist()
        for k, v in itertools.groupby(enumerate(steps), key=lambda x: x[1]):
            v = sorted(v, key=lambda x: x[0])
            step_indices.append(v[-1][0])
        data = data.transpose()[1:, step_indices[:nframes]]
        #config._print(data)

        for i, atoms in enumerate(curr_traj_frames):
            for j, k in enumerate(dkeys):
                try:
                    atoms.info[k] = data[j,i]
                except IndexError:
                    # NOTE: Some potentials donot print last frames of min
                    #       for example, lammps
                    atoms.info[k] = 0.
    else:
        ...
    
    # - check COLVAR
    if colvar_io is not None:
        # - read latest COLVAR Files
        names = colvar_io.readline().split()[2:]
        colvar_io.seek(0)
        colvars = np.loadtxt(colvar_io)
        #print("colvars: ", colvars.shape)
        curr_colvars = colvars[-nframes_traj:, :]
        for i, atoms in enumerate(curr_traj_frames):
            for k, v in zip(names, curr_colvars[i, :]):
                atoms.info[k] = v
    
    # - Close IO
    traj_io.close()
    log_io.close()
    if prism_io is not None:
        prism_io.close()
    if devi_io is not None:
        devi_io.close()
    if colvar_io is not None:
        colvar_io.close()

    return curr_traj_frames

@dataclasses.dataclass
class LmpDriverSetting(DriverSetting):

    min_style: str = "fire"
    min_modify: str = "integrator verlet tmax 4"

    etol: float = 0
    fmax: float = 0.05

    neighbor: str = "0.0 bin"
    neigh_modify: str = None

    extra_fix: List[str] = dataclasses.field(default_factory=list)

    plumed: str = None

    def __post_init__(self):
        """"""
        if self.task == "min":
            self._internals.update(
                min_style = self.min_style,
                min_modify = self.min_modify,
                etol = self.etol,
                ftol = self.fmax,
                #maxstep = self.maxstep
            )
        
        if self.task == "md":
            self._internals.update(
                md_style = self.md_style,
                timestep = self.timestep,
                velocity_seed = self.velocity_seed,
                ignore_atoms_velocities = self.ignore_atoms_velocities,
                remove_rotation = self.remove_rotation,
                remove_translation = self.remove_translation,
                temp = self.temp,
                # TODO: end temperature
                Tdamp = self.Tdamp,
                press = self.press,
                Pdamp = self.Pdamp,
                # - ext
                plumed = self.plumed,
            )
        
        # - shared params
        self._internals.update(
            task = self.task,
            dump_period = self.dump_period,
            ckpt_period = self.ckpt_period
        )

        # - special params
        self._internals.update(
            neighbor = self.neighbor,
            neigh_modify = self.neigh_modify,
            extra_fix = self.extra_fix
        )

        return 
    
    def get_run_params(self, *args, **kwargs):
        """"""
        # - pop out special keywords
        # convergence criteria
        ftol_ = kwargs.pop("fmax", self.fmax)
        etol_ = kwargs.pop("etol", self.etol)
        if etol_ is None:
            etol_ = 0.
        if ftol_ is None:
            ftol_ = 0.

        steps_ = kwargs.pop("steps", self.steps)

        run_params = dict(
            constraint = kwargs.get("constraint", self.constraint),
            etol=etol_, ftol=ftol_, maxiter=steps_, maxeval=2*steps_
        )

        # - add extra parameters
        run_params.update(
            **kwargs
        )

        return run_params

class LmpDriver(AbstractDriver):

    """Use lammps to perform dynamics.
    
    Minimisation and/or molecular dynamics.

    """

    name = "lammps"

    special_keywords = {}

    default_task = "min"
    supported_tasks = ["min", "md"]

    #: List of output files would be saved when restart.
    saved_fnames: List[str] = [ASELMPCONFIG.log_filename, ASELMPCONFIG.trajectory_filename, ASELMPCONFIG.deviation_filename]

    def __init__(self, calc, params: dict, directory="./", *args, **kwargs):
        """"""
        calc, params = self._check_plumed(calc=calc, params=params)

        super().__init__(calc, params, directory=directory, *args, **kwargs)
        self.setting = LmpDriverSetting(**params)

        return
    
    def _check_plumed(self, calc, params: dict):
        """"""
        new_calc, new_params = calc, params
        if isinstance(calc, LinearCombinationCalculator):
            ncalcs = len(calc.calcs)
            assert ncalcs == 2, "Number of calculators should be 2."
            if isinstance(calc.calcs[0], Lammps) and isinstance(calc.calcs[1], Plumed):
                new_calc = calc.calcs[0]
                new_params = copy.deepcopy(params)
                new_params["plumed"] = "".join(calc.calcs[1].input)

        return new_calc, new_params
    
    def _verify_checkpoint(self, *args, **kwargs) -> bool:
        """"""
        verified = super()._verify_checkpoint(*args, **kwargs)
        if verified:
            checkpoints = list(self.directory.glob("restart.*"))
            self._debug(f"checkpoints: {checkpoints}")
            if not checkpoints:
                verified = False
        else:
            ...

        return verified

    def _irun(self, atoms: Atoms, ckpt_wdir=None, *args, **kwargs):
        """"""
        try:
            # - params
            run_params = self.setting.get_init_params()
            run_params.update(**self.setting.get_run_params(**kwargs))

            if ckpt_wdir is None: # start from the scratch
                ...
            else:
                checkpoints = sorted(list(ckpt_wdir.glob("restart.*")), key=lambda x: int(x.name.split(".")[1]))
                self._debug(f"checkpoints to restart: {checkpoints}")
                target_steps = run_params["maxiter"]
                run_params.update(
                    read_restart = str(checkpoints[-1].resolve()),
                    maxiter = target_steps - int(checkpoints[-1].name.split(".")[1])
                )

            # - check constraint
            self.calc.set(**run_params)
            atoms.calc = self.calc

            # - run
            _ = atoms.get_forces()
        except Exception as e:
            config._debug(e)

        return
    
    def read_trajectory(
            self, type_list=None, add_step_info=True, 
            archive_path: pathlib.Path=None, *args, **kwargs
        ) -> List[Atoms]:
        """Read trajectory in the current working directory."""
        if type_list is not None:
            self.calc.type_list = type_list
        curr_units = self.calc.units

        # - find runs...
        prev_wdirs = sorted(self.directory.glob(r"[0-9][0-9][0-9][0-9][.]run"))
        self._debug(f"prev_wdirs: {prev_wdirs}")

        traj_list = []
        for w in prev_wdirs:
            curr_frames = read_single_simulation(
                mdir=self.directory, wdir=w, prefix="", units=curr_units, 
                add_step_info=add_step_info, archive_path=archive_path
            )
            traj_list.append(curr_frames)
        
        # Even though traj file may be empty, the read can give a empty list...
        traj_list.append(
            read_single_simulation(
                mdir=self.directory, wdir=self.directory, prefix="", units=curr_units, 
                add_step_info=add_step_info, archive_path=archive_path
            )
        )

        # -- concatenate
        traj_frames, ntrajs = [], len(traj_list)
        if ntrajs > 0:
            traj_frames.extend(traj_list[0])
            for i in range(1, ntrajs):
                assert np.allclose(traj_list[i-1][-1].positions, traj_list[i][0].positions), f"Traj {i-1} and traj {i} are not consecutive in positions."
                assert np.allclose(traj_list[i-1][-1].get_potential_energy(), traj_list[i][0].get_potential_energy()), f"Traj {i-1} and traj {i} are not consecutive in energy."
                traj_frames.extend(traj_list[i][1:])
        else:
            ...

        return traj_frames
    

class Lammps(FileIOCalculator):

    #: Calculator name.
    name: str = "Lammps"

    #: Implemented properties.
    implemented_properties: List[str] = ["energy", "forces", "stress"]

    #: LAMMPS command.
    command: str = "lmp 2>&1 > lmp.out"

    #: Default calculator parameters, NOTE which have ase units.
    default_parameters: dict = dict(
        # ase params
        task = "min",
        constraint = None, # index of atoms, start from 0
        ignore_atoms_velocities = False,
        # --- lmp params ---
        read_restart = None,
        units = "metal",
        atom_style = "atomic",
        processors = "* * 1",
        #boundary = "p p p",
        newton = None,
        pair_style = None,
        pair_coeff = None,
        neighbor = "0.0 bin",
        neigh_modify = None,
        mass = "* 1.0",
        dump_period = 1,
        ckpt_period = 100,
        # - md
        md_style = "nvt",
        md_steps = 0,
        velocity_seed = None,
        timestep = 1.0, # fs
        temp = 300,
        pres = 1.0,
        Tdamp = 100, # fs
        Pdamp = 100,
        # - minimisation
        etol = 0.0,
        ftol = 0.05,
        maxiter = 0, # NOTE: this is steps for MD
        maxeval = 0,
        min_style = "fire",
        min_modify = "integrator verlet tmax 4",
        # - extra fix
        extra_fix = [],
        # - externals
        plumed = None
    )

    #: Symbol to integer.
    type_list: List[str] = None

    #: Cached trajectory of the previous simulation.
    cached_traj_frames: List[Atoms] = None

    def __init__(
        self, 
        command = None, 
        label = name, 
        **kwargs
    ):
        """"""
        FileIOCalculator.__init__(self, command=command, label=label, **kwargs)

        # - check potential
        assert self.pair_style is not None, "pair_style is not set."

        return
    
    def __getattr__(self, key):
        """Corresponding getattribute-function."""
        if key != "parameters" and key in self.parameters:
            return self.parameters[key]
        return object.__getattribute__(self, key)
    
    def calculate(self, atoms=None, properties=["energy"],
            system_changes=all_changes): 
        """Run calculation."""
        # TODO: should use user-custom type_list from potential manager
        #       move this part to driver?
        self.type_list = parse_type_list(atoms)

        # init for creating the directory
        FileIOCalculator.calculate(self, atoms, properties, system_changes)

        return
    
    def write_input(self, atoms, properties=None, system_changes=None) -> None:
        """Write input file and input structure."""
        FileIOCalculator.write_input(self, atoms, properties, system_changes)

        # - check velocities
        self.write_velocities = False
        if atoms.get_kinetic_energy() > 0.:
            self.write_velocities = (True and not self.ignore_atoms_velocities)

        # write structure
        prismobj = Prism(atoms.get_cell()) # TODO: nonpbc?
        prism_file = os.path.join(self.directory, ASELMPCONFIG.prism_filename)
        with open(prism_file, "wb") as fopen:
            pickle.dump(prismobj, fopen)
        stru_data = os.path.join(self.directory, ASELMPCONFIG.inputstructure_filename)
        write_lammps_data(
            stru_data, atoms, specorder=self.type_list, 
            force_skew=True, prismobj=prismobj, velocities=self.write_velocities,
            units=self.units, atom_style=self.atom_style
        )

        # write input
        self._write_input(atoms)

        return
    
    def _is_finished(self):
        """Check whether the simulation finished or failed. 

        Return wall time if the simulation finished.

        """

        is_finished, end_info = False, "not finished"
        log_filepath = pathlib.Path(os.path.join(self.directory, ASELMPCONFIG.log_filename))

        if log_filepath.exists():
            ERR_FLAG = "ERROR: "
            END_FLAG = "Total wall time:"
            with open(log_filepath, "r") as fopen:
                lines = fopen.readlines()
        
            for line in lines:
                if line.strip().startswith(ERR_FLAG):
                    is_finished = True
                    end_info = " ".join(line.strip().split()[1:])
                    break
                if line.strip().startswith(END_FLAG):
                    is_finished = True
                    end_info = " ".join(line.strip().split()[1:])
                    break
            else:
                is_finished = False
        else:
            is_finished = False

        return is_finished, end_info 
    
    def read_results(self):
        """ASE read results."""
        # obtain results
        self.results = {}

        # - Be careful with UNITS
        # read forces from dump file
        curr_wdir = pathlib.Path(self.directory)
        self.cached_traj_frames = read_single_simulation(
            mdir=curr_wdir, wdir=curr_wdir, prefix="", 
            units=self.units, add_step_info=True
        )
        converged_frame = self.cached_traj_frames[-1]

        self.results["forces"] = converged_frame.get_forces().copy()
        self.results["energy"] = converged_frame.get_potential_energy()

        # - add deviation info
        for k, v in converged_frame.info.items():
            if "devi" in k:
                self.results[k] = v

        return

    def _write_input(self, atoms) -> None:
        """Write input file in.lammps"""
        # - write in.lammps
        content =  f"restart         {self.ckpt_period}  restart.*.data\n\n"
        content += "units           %s\n" %self.units
        content += "atom_style      %s\n" %self.atom_style

        # - mpi settings
        if self.processors is not None:
            content += "processors {}\n".format(self.processors) # if 2D simulation
        
        # - simulation box
        pbc = atoms.get_pbc()
        if "boundary" in self.parameters:
            content += "boundary {0} \n".format(self.parameters["boundary"])
        else:
            content += "boundary {0} {1} {2} \n".format(
                *tuple("fp"[int(x)] for x in pbc) # sometimes s failed to wrap all atoms
            )
        content += "\n"
        if self.newton:
            content += "newton {}\n".format(self.newton)
        content += "box             tilt large\n"
        if self.read_restart is None:
            content += "read_data	    %s\n" %ASELMPCONFIG.inputstructure_filename
        else:
            content += f"read_restart    {self.read_restart}\n"
            #os.remove(ASELMPCONFIG.inputstructure_filename)
        content += "change_box      all triclinic\n"

        # - particle masses
        mass_line = "".join(
            "mass %d %f\n" %(idx+1,atomic_masses[atomic_numbers[elem]]) for idx, elem in enumerate(self.type_list)
        )
        content += mass_line
        content += "\n"

        # - pair, MLIP specific settings
        # TODO: neigh settings?
        potential = self.pair_style.strip().split()[0]
        if potential == "reax/c":
            assert self.atom_style == "charge", "reax/c should have charge atom_style"
            content += "pair_style  {}\n".format(self.pair_style)
            content += "pair_coeff {} {}\n".format(self.pair_coeff, " ".join(self.type_list))
            content += "fix             reaxqeq all qeq/reax 1 0.0 10.0 1e-6 reax/c\n"
        elif potential == "eann":
            pot_data = self.pair_style.strip().split()[1:]
            endp = len(pot_data)
            for ip, p in enumerate(pot_data):
                if p == "out_freq":
                    endp = ip
                    break
            pot_data = pot_data[:endp]
            if len(pot_data) > 1:
                pair_style = "eann {} out_freq {}".format(" ".join(pot_data), self.dump_period)
            else:
                pair_style = "eann {}".format(" ".join(pot_data))
            content += "pair_style  {}\n".format(pair_style)
            # NOTE: make out_freq consistent with dump_period
            if self.pair_coeff is None:
                pair_coeff = "double * *"
            else:
                pair_coeff = self.pair_coeff
            content += "pair_coeff	{} {}\n".format(pair_coeff, " ".join(self.type_list))
        elif potential == "deepmd":
            content += "pair_style  {} out_freq {}\n".format(self.pair_style, self.dump_period)
            content += "pair_coeff	{} {}\n".format(self.pair_coeff, " ".join(self.type_list))
        else:
            content += "pair_style {}\n".format(self.pair_style)
            #content += "pair_coeff {} {}\n".format(self.pair_coeff, " ".join(self.type_list))
            content += "pair_coeff {}\n".format(self.pair_coeff)
        content += "\n"

        # - neighbor
        content += "neighbor        {}\n".format(self.neighbor)
        if self.neigh_modify:
            content += "neigh_modify        {}\n".format(self.neigh_modify)
        content += "\n"

        # - constraint
        mobile_text, frozen_text = parse_constraint_info(atoms, self.constraint)
        if mobile_text: # NOTE: sometimes all atoms are fixed
            content += "group mobile id %s\n" %mobile_text
            content += "\n"
        if frozen_text: # not empty string
            # content += "region bottom block INF INF INF INF 0.0 %f\n" %zmin # unit A
            content += "group frozen id %s\n" %frozen_text
            content += "fix cons frozen setforce 0.0 0.0 0.0\n"
        content += "\n"

        # - outputs
        # TODO: use more flexible notations
        if self.task == "min":
            content += "thermo_style    custom step pe ke etotal temp press vol fmax fnorm\n"
        elif self.task == "md":
            content += "compute mobileTemp mobile temp\n"
            content += "thermo_style    custom step c_mobileTemp pe ke etotal press vol lx ly lz xy xz yz\n"
        else:
            pass
        content += "thermo          {}\n".format(self.dump_period) 

        # TODO: How to dump total energy?
        content += "dump		1 all custom {} {} id type element x y z fx fy fz vx vy vz\n".format(
            self.dump_period, ASELMPCONFIG.trajectory_filename
        )
        content += "dump_modify 1 element {}\n".format(" ".join(self.type_list))
        content += "\n"

        # - add extra fix
        for i, fix_info in enumerate(self.extra_fix):
            content += "{:<24s}  {:<24s}  {:<s}\n".format("fix", f"extra{i}", fix_info)
        
        # --- run type
        if self.task == "min":
            # - minimisation
            content += "min_style       {}\n".format(self.min_style)
            content += "min_modify      {}\n".format(self.min_modify)
            content += "minimize        {:f} {:f} {:d} {:d}\n".format(
                unitconvert.convert(self.etol, "energy", "ASE", self.units),
                unitconvert.convert(self.ftol, "force", "ASE", self.units),
                self.maxiter, self.maxeval
            )
        elif self.task == "md":
            if not self.write_velocities and self.read_restart is None:
                velocity_seed = self.velocity_seed
                if velocity_seed is None:
                    velocity_seed = np.random.randint(0,10000)
                velocity_command = "velocity        mobile create {} {} dist gaussian ".format(self.temp, velocity_seed)
                if hasattr(self, "remove_translation"):
                    if self.remove_translation:
                        velocity_command += "mom yes "
                if hasattr(self, "remove_rotation"):
                    if self.remove_rotation:
                        velocity_command += "rot yes "
                velocity_command += "\n"
                content += velocity_command
        
            if self.md_style == "nvt":
                Tdamp_ = unitconvert.convert(self.Tdamp, "time", "real", self.units)
                content += "fix             thermostat mobile nvt temp {} {} {}\n".format(
                    self.temp, self.temp, Tdamp_
                )
            elif self.md_style == "npt":
                pres_ = unitconvert.convert(self.pres, "pressure", "metal", self.units)
                Tdamp_ = unitconvert.convert(self.Tdamp, "time", "real", self.units)
                Pdamp_ = unitconvert.convert(self.Pdamp, "time", "real", self.units)
                content += "fix             thermostat mobile npt temp {} {} {} aniso {} {} {}\n".format(
                    self.temp, self.temp, Tdamp_, pres_, pres_, Pdamp_
                )
            elif self.md_style == "nve":
                content += "fix             thermostat mobile nve \n"

            timestep_ = unitconvert.convert(self.timestep, "time", "real", self.units)
            content += "\n"
            content += f"timestep        {timestep_}\n"
            if self.plumed is not None:
                plumed_inp = update_stride_and_file(self.plumed, wdir=str(self.directory), stride=self.dump_period)
                with open(os.path.join(self.directory, "plumed.inp"), "w") as fopen:
                    fopen.write("".join(plumed_inp))
                content += "fix             metad all plumed plumedfile plumed.inp outfile plumed.out\n"
            content += f"run             {self.maxiter}\n"
        else:
            # TODO: NEB?
            ...
    
        # - output file
        in_file = os.path.join(self.directory, ASELMPCONFIG.input_fname)
        with open(in_file, "w") as fopen:
            fopen.write(content)

        return
 

if __name__ == "__main__":
    ...



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import abc
import copy
from typing import NoReturn, Callable, List

import numpy as np

from ase import Atoms
from ase import data, units

from .. import registers


class AbstractOperator(abc.ABC):

    #: Operator name.
    name: str = "abstract"

    #: Maximum attempts to generate the new position of a species.
    MAX_RANDOM_ATTEMPTS = 1000

    #: Print function.
    _print: Callable = print

    def __init__(
        self,
        region: dict = {},
        temperature: float = 300.0,
        pressure: float = 1.0,
        covalent_ratio=[0.8, 2.0],
        use_rotation=True,
        prob: str = 1.0,
        allow_isolated: bool = False,
        *args,
        **kwargs,
    ) -> None:
        """Initialise the modification operator.

        Args:
            temperature: Monte Carlo temperature [K].
            pressure: Monte Carlo pressure [bar].
            covalent_ratio: Minimum and maximum percentages of covalent distance.

        """
        super().__init__()

        # - region
        region_params = copy.deepcopy(region)
        region_method = region_params.pop("method", "auto")
        self.region = registers.create(
            "region", region_method, convert_name=True, **region_params
        )

        # - thermostat
        self.temperature = temperature
        self.pressure = pressure

        # - restraint on atomic distance
        self.covalent_min = covalent_ratio[0]
        self.covalent_max = covalent_ratio[1]

        # - molecule
        self.use_rotation = use_rotation

        # - probability
        self.prob = prob

        # - neighbour setting
        self.allow_isolated = allow_isolated

        return

    def _check_region(self, atoms: Atoms, *args, **kwargs):
        """"""
        if self.region.__class__.__name__ == "AutoRegion":
            self.region._curr_atoms = atoms

        # NOTE: Modify only atoms in the region...
        tags_dict = self.region.get_tags_dict(atoms)
        content = "species within system:\n"
        content += (
            "  "
            + "  ".join([str(k) + " " + str(len(v)) for k, v in tags_dict.items()])
            + "\n"
        )
        for x in content.split("\n"):
            self._print(x)

        self._curr_tags_dict = self.region.get_contained_tags_dict(atoms, tags_dict)
        content = "species within region:\n"
        content += (
            "  "
            + "  ".join(
                [str(k) + " " + str(len(v)) for k, v in self._curr_tags_dict.items()]
            )
            + "\n"
        )
        for x in content.split("\n"):
            self._print(x)

        return

    def _select_species(
        self, atoms: Atoms, particles: List[str] = None, rng=np.random
    ) -> List[int]:
        """"""
        # - pick a particle (atom/molecule)
        tags_within_region = []
        for k, v in self._curr_tags_dict.items():
            if particles is not None and k not in particles:
                continue
            tags_within_region.extend(v)
        self._print(f"ntags of {particles}: {len(tags_within_region)}")

        if len(tags_within_region) > 0:
            picked_tag = rng.choice(tags_within_region)
            tags = atoms.get_tags()
            species_indices = [i for i, t in enumerate(tags) if t == picked_tag]
            self._print(
                f"selected tag: {picked_tag} species: {atoms[species_indices].get_chemical_formula()}"
            )
        else:
            picked_tag = None
            species_indices = None
            raise RuntimeError(f"{self.__class__.__name__} does not have {particles}.")

        return species_indices

    def _rotate_species(self, species: Atoms, rng=np.random):
        """"""
        species_ = species.copy()  # TODO: make clean atoms?
        org_com = np.mean(species_.positions, axis=0)
        if self.use_rotation and len(species_) > 1:
            phi, theta, psi = 360 * rng.uniform(0, 1, 3)
            species_.euler_rotate(phi=phi, theta=0.5 * theta, psi=psi, center=org_com)

        return species_

    def check_overlap_neighbour(
        self, nl, new_atoms: Atoms, cell, species_indices: List[int]
    ) -> bool:
        """Check whether the species position is valid.

        Use neighbour list to check newly added atom is neither too close or too
        far from other atoms. The neighbour list is based on covalent_max distance.
        We have three status, `valid`, `invalid`, adn `isolated`.
        The situations being considered invalid:
            - Atomic distance too close (covalent_min).
            - All atoms in the species are isolated from the rest of the system.

        """
        assert self.blmin is not None, "BondLengthMinimumDict is not properly set."

        # -
        num_atoms_in_species = len(species_indices)

        species_status = ["valid"] * num_atoms_in_species
        self._print(f"- {species_indices =}")

        # - get symbols here since some operators may change the symbol
        chemical_symbols = new_atoms.get_chemical_symbols()

        nl.update(new_atoms)
        for iatom, idx_pick in enumerate(species_indices):
            indices, offsets = nl.get_neighbors(idx_pick)
            self._debug(
                f"  check index {idx_pick} {new_atoms.positions[idx_pick]} nneighs: {len(indices)}"
            )
            if len(indices) > 0:
                # --
                if all([(ni in species_indices) for ni in indices]):
                    species_status[iatom] = "isolated"
                    continue
                # -- check inter-species atomic distances
                for ni, offset in zip(indices, offsets):
                    # NOTE: Check if the species contact other atoms
                    #       in a reasonable distance.
                    #       Intra-species distance will not be checked.
                    if ni not in species_indices:
                        dis = np.linalg.norm(
                            new_atoms.positions[idx_pick]
                            - (new_atoms.positions[ni] + np.dot(offset, cell))
                        )
                        pairs = [chemical_symbols[ni], chemical_symbols[idx_pick]]
                        pairs = tuple([data.atomic_numbers[p] for p in pairs])
                        if dis <= self.blmin[pairs]:
                            species_status[iatom] = "invalid"
                            self._debug(f"  distance: {ni} {dis} {self.blmin[pairs]}")
                            break
                    else:
                        ...
                else:
                    species_status[iatom] = "valid"
            else:
                # We need BREAK here?
                # For a single-atom species, one loop finishes.
                # For a multi-atom species, the code will not happen except when
                # the covalent_max is WAY TOO SMALL then ...
                # assert num_atoms_in_species == 1,
                species_status[iatom] = "isolated"
        self._print(f"  {species_status =}")

        status_ = []
        for s in species_status:
            if s == "valid" or s == "invalid":
                ...
            else:  # isolated
                if self.allow_isolated:
                    s = "valid"
                else:
                    s = "invalid"
            status_.append(s)

        return any([s == "invalid" for s in status_])

    @abc.abstractmethod
    def run(self, atoms: Atoms, rng=np.random) -> Atoms:
        """Modify the input atoms.

        Returns:
            A new atoms is returned if the modification succeeds otherwise None is
            returned.

        """
        self._check_region(atoms)

        return

    @abc.abstractmethod
    def metropolis(self, prev_ene: float, curr_ene: float, rng=np.random) -> bool:
        """Monte Carlo."""

        return

    def as_dict(self) -> dict:
        """"""
        params = {}
        params["method"] = self.name
        params["region"] = self.region.as_dict()
        params["temperature"] = self.temperature
        params["pressure"] = self.pressure
        params["covalent_ratio"] = [self.covalent_min, self.covalent_max]
        params["use_rotation"] = self.use_rotation
        params["prob"] = self.prob

        params = copy.deepcopy(params)

        return params


if __name__ == "__main__":
    ...



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


import pathlib
from typing import List

import numpy as np

from ase.calculators.calculator import Calculator
from ase.geometry import find_mic


def compute_rmsd_energy_and_forces(
    cell,
    coordinate,
    saved_coordinates,
    damped_steps: List[int],
    kappa: float,
    sigma: float,
    omega: float,
    pbc: bool = True,
):
    """"""
    # - preprocess coordinates
    num_atoms = len(coordinate)
    num_records = len(saved_coordinates)

    vectors, rmsd2 = [], []  # rmsd square
    if pbc:
        for i in range(num_records):
            mic_vec, mic_dis = find_mic(coordinate - saved_coordinates[i], cell)
            vectors.append(mic_vec)
            rmsd2_i = np.sum(mic_vec**2) / num_atoms
            rmsd2.append(rmsd2_i)
    else:
        for i in range(num_records):
            mic_vec = coordinate - saved_coordinates[i]
            vectors.append(mic_vec)
            rmsd2_i = np.sum(mic_vec**2) / num_atoms
            rmsd2.append(rmsd2_i)
    vectors = np.array(vectors)
    rmsd = np.sqrt(np.array(rmsd2))[:, np.newaxis]  # shape (num_references, 1)

    # - compute energy
    s1 = rmsd
    s2 = s1**2 / 2.0 / sigma**2  # uniform sigma?
    v = omega * np.exp(-np.sum(s2, axis=1))

    energy = v.sum(axis=0)

    # - compute forces
    # -- dE/ds _ shape (num_references, 1)
    dEds = -v[:, np.newaxis] * s1 / sigma**2

    # -- ds/dc # cv gradient wrt cartesian coordinate
    dsdc = (
        (1.0 / (num_atoms * rmsd))[:, :, np.newaxis]
        .repeat(num_atoms, axis=1)
        .repeat(3, axis=2)
    ) * vectors  # shape (num_references, num_atoms, 3)

    damped_prefactors = compute_damping_function(damped_steps, kappa=kappa)[
        :, np.newaxis, np.newaxis
    ]

    forces = -np.sum(damped_prefactors * dEds[:, :, np.newaxis] * dsdc, axis=0)

    return energy, forces


def compute_damping_function(steps, kappa: float = 0.03):
    """"""
    damp = 2.0 / (1.0 + np.exp(-kappa * steps)) - 1.0

    return damp


class RMSDGaussian(Calculator):

    implemented_properties = ["energy", "fee_energy", "forces"]

    def __init__(
        self,
        group: List[int],
        pace: int = 1,
        sigma: float = 0.2,
        omega: float = 0.5,
        kappa: float = 0.03,
        memory_length: int = 0,
        **kwargs,
    ):
        """"""
        super().__init__(**kwargs)

        self.group = group

        self.pace = pace

        self.sigma = sigma

        self.omega = omega

        self.kappa = kappa

        self.memory_length = memory_length

        # - private
        self._history_records = []
        self._history_steps = []

        self._num_steps = 0

        return

    @property
    def num_steps(self) -> int:
        """"""

        return self._num_steps

    def calculate(
        self,
        atoms=None,
        properties=["energy"],
        system_changes=["positions", "numbers", "cell"],
    ):
        super().calculate(atoms, properties, system_changes)

        if self.num_steps == 0:
            content = f"# pace {self.pace} width {self.sigma} height {self.omega}\n"
            content += f"# memory {self.memory_length} kappa {self.kappa}\n"
            content += "# {:>10s}  {:>12s}  {:>12s}\n".format(
                "step", "num_records", "energy"
            )

            log_fpath = pathlib.Path(self.directory) / "info.log"
            with open(log_fpath, "w") as fopen:
                fopen.write(content)

        coordinate = atoms.get_positions()[self.group]

        # Remove previous references
        self._history_records = self._history_records[-self.memory_length :]
        self._history_steps = self._history_steps[-self.memory_length :]

        # The same coordinate leads to zero division so we use [:-1] records
        # so we add record after computing energy and forces.
        saved_coordinates = np.array(self._history_records)
        damped_steps = self.num_steps - np.array(self._history_steps)
        if saved_coordinates.shape[0] == 0:
            energy = 0.0
            forces = np.zeros(atoms.positions.shape)
        else:
            energy, biased_forces = compute_rmsd_energy_and_forces(
                cell=atoms.get_cell(complete=True),
                coordinate=coordinate,
                saved_coordinates=saved_coordinates,
                damped_steps=damped_steps,
                kappa=self.kappa,
                sigma=self.sigma,
                omega=self.omega,
                pbc=True,  # FIXME: nopbc systems?
            )
            forces = np.zeros(atoms.positions.shape)
            forces[self.group] = biased_forces

        self.results["energy"] = energy
        self.results["free_energy"] = energy
        self.results["forces"] = forces

        self._write_step(energy)

        if self.num_steps % self.pace == 0:
            self._history_records.append(coordinate)
            self._history_steps.append(self.num_steps)

        self._num_steps += 1

        return

    def _write_step(self, energy: float):
        """"""
        num_records = len(self._history_records)
        content = "{:>12d}  {:>12d}  {:>12.4f}\n".format(
            self.num_steps, num_records, energy
        )

        log_fpath = pathlib.Path(self.directory) / "info.log"
        with open(log_fpath, "a") as fopen:
            fopen.write(content)

        return


if __name__ == "__main__":
    ...



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import copy
from typing import Optional, List

import numpy as np

from ase import Atoms

from .builder import StructureModifier 
from .utils import str2list_int, check_overlap_neighbour


class PerturbatorBuilder(StructureModifier):

    name = "perturbater"

    #: Number of attempts to create a random candidate.
    MAX_ATTEMPTS_PER_CANDIDATE: int = 1000

    #: Number of attempts to create a number of candidates.
    #       if 10 structures are to create, run will try 5*10=50 times.
    MAX_TIMES_SIZE: int = 5

    """Perturb positions of input structures.

    TODO:
        1. Perturb distances, angles...

    """

    def __init__(
            self, eps: Optional[float]=None, ceps: Optional[float]=None, 
            group: Optional[str]=None,
            covalent_ratio=[0.8, 2.0],
            max_times_size: int = 5, *args, **kwargs
        ):
        """Initialise a perturbator.

        Args:
            eps: Drift on atomic positions. Unit Ang.
            ceps: Drift on lattice constants. Unit Ang.

        """
        super().__init__(*args, **kwargs)

        # - check inputs
        if eps is None and ceps is None:
            raise RuntimeError("Either eps or ceps should be provided.")
        elif eps is not None and ceps is None:
            ...
        elif eps is None and ceps is not None:
            ...
        else:
            ...

        self.eps = eps
        self.ceps = ceps

        # - apply perturbation on a group of atoms, default is all
        self.group = group
        if self.group is not None:
            self.group = str2list_int(self.group, convention="lmp")

        # - distance check
        self.covalent_ratio = covalent_ratio
        self.MAX_TIMES_SIZE = max_times_size

        return
    
    def run(self, substrates: Optional[List[Atoms]]=None, size:int=1, *args, **kwargs) -> List[Atoms]:
        """"""
        super().run(substrates=substrates, *args, **kwargs)

        frames = []
        for substrate in self.substrates:
            curr_frames = self._irun(substrate, size)
            frames.extend(curr_frames)

        return frames
    
    def _irun(self, substrate: Atoms, size: int):
        """"""
        frames = []
        for i in range(size*self.MAX_TIMES_SIZE):
            nframes = len(frames)
            if nframes < size:
                atoms = copy.deepcopy(substrate)
                if self.eps is not None:
                    natoms = len(atoms)
                    pos_drift = self.rng.random((natoms, 3))
                    if self.group is not None:
                        pos_drift_ = np.zeros((natoms, 3))
                        pos_drift_[self.group] = pos_drift[self.group]
                        pos_drift = pos_drift_
                    atoms.positions += pos_drift*self.eps
                if all(atoms.pbc) and self.ceps is not None:
                    lat_drift = self.rng.random((3, 3))
                    atoms.cell += lat_drift*self.ceps
                if check_overlap_neighbour(atoms, self.covalent_ratio):
                    frames.append(atoms)
            else:
                break
        else:
            nframes = len(frames)
            raise RuntimeError(
                f"Failed to create {size} structures, only {nframes} are created."
            )

        return frames


if __name__ == "__main__":
    ...



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


import copy

from ase import Atoms


class ScfErrAtoms(Atoms):

    def __init__(
        self,
        symbols=None,
        positions=None,
        numbers=None,
        tags=None,
        momenta=None,
        masses=None,
        magmoms=None,
        charges=None,
        scaled_positions=None,
        cell=None,
        pbc=None,
        celldisp=None,
        constraint=None,
        calculator=None,
        info=None,
        velocities=None,
    ):
        super().__init__(
            symbols,
            positions,
            numbers,
            tags,
            momenta,
            masses,
            magmoms,
            charges,
            scaled_positions,
            cell,
            pbc,
            celldisp,
            constraint,
            calculator,
            info,
            velocities,
        )

        self.info["scf_error"] = True

        return

    @staticmethod
    def from_atoms(atoms: Atoms) -> "ScfErrAtoms":
        """"""
        scferr_atoms = ScfErrAtoms(
            symbols=atoms.get_chemical_symbols(),
            positions=atoms.get_positions(),
            # tags=atoms.get_tags(),
            # magmoms
            # charges
            momenta=atoms.get_momenta(),
            cell=atoms.get_cell(),
            pbc=atoms.get_pbc(),
            constraint=atoms.constraints,
            calculator=atoms.calc,
            info=copy.deepcopy(atoms.info),
        )

        return scferr_atoms
    

if __name__ == "__main__":
    ...



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import abc
import copy
import pathlib
from typing import NoReturn, Optional, Union, Callable, List, Tuple

import numpy as np

from tinydb import TinyDB, Query

from .. import config
from ..scheduler.scheduler import AbstractScheduler


"""worker = driver + scheduler.

A worker that manages a series of dynamics tasks
worker needs a scheduler to dertermine whether run by serial
or on cluster.

"""


class AbstractWorker(abc.ABC):
    """"""

    UUIDLEN = 36  # length of uuid

    _print: Callable = config._print
    _debug: Callable = config._debug

    # Number of executation in each queue job.
    batchsize: int = 1

    _directory: Optional[pathlib.Path] = None
    _scheduler = None
    _database = None

    _submit = True

    _script_name = "run.script"

    def __init__(
        self,
        directory: Optional[Union[str, pathlib.Path]] = None,
        batchsize=1,
        *args,
        **kwargs,
    ) -> None:
        """ """
        # - set default directory
        if directory is not None:
            self.directory = directory

        self.batchsize = batchsize

        self.n_jobs = config.NJOBS

        return

    @property
    def directory(self) -> pathlib.Path:
        """"""

        return self._directory

    @directory.setter
    def directory(self, directory_: Union[str, pathlib.Path]):
        self._directory = pathlib.Path(directory_)

        return

    @property
    def scheduler(self):

        return self._scheduler

    @scheduler.setter
    def scheduler(self, scheduler_):
        """"""
        assert isinstance(scheduler_, AbstractScheduler), ""
        self._scheduler = scheduler_

        # - update mode
        # if self._scheduler.name == "local"

        return

    def _split_groups(self, npoints: int) -> Tuple[List[int], List[int]]:
        """Split nframes into groups."""
        # - split frames
        ngroups = int(np.floor(1.0 * npoints / self.batchsize))
        group_indices = [0]
        for i in range(ngroups):
            group_indices.append((i + 1) * self.batchsize)
        if group_indices[-1] != npoints:
            group_indices.append(npoints)
        starts, ends = group_indices[:-1], group_indices[1:]
        assert len(starts) == len(ends), "Inconsistent start and end indices..."
        # group_indices = [f"{s}:{e}" for s, e in zip(starts,ends)]

        return (starts, ends)

    # @property
    # def database(self):

    #    return self._database

    # @database.setter
    # def database(self, database_):
    #    self._database = database_

    #    return

    def _initialise(self, *args, **kwargs):
        """"""
        if not self.directory.exists():
            self.directory.mkdir(parents=True, exist_ok=True)  # NOTE: ./tmp_folder
        else:
            ...
        assert self.directory, "Working directory is not set properly..."

        return

    @abc.abstractmethod
    def run(self, *args, **kwargs):
        """"""
        self._initialise(*args, **kwargs)
        # - some imported packages change `logging.basicConfig`
        #   and accidently add a StreamHandler to logging.root
        #   so remove it...
        # for h in logging.root.handlers:
        #    if isinstance(h, logging.StreamHandler) and not isinstance(h, logging.FileHandler):
        #        logging.root.removeHandler(h)
        self._print(f"~~~{self.__class__.__name__}+run")
        return

    def inspect(self, resubmit=False, *args, **kwargs):
        """"""
        self._initialise(*args, **kwargs)
        self._debug(f"~~~{self.__class__.__name__}+inspect")

        running_jobs = self._get_running_jobs()

        return

    def retrieve(self, *args, **kwargs):
        """"""
        self.inspect(*args, **kwargs)
        self._print(f"~~~{self.__class__.__name__}+retrieve")

        results = []

        return results

    def _get_running_jobs(self):
        """"""
        with TinyDB(
            self.directory / f"_{self.scheduler.name}_jobs.json", indent=2
        ) as database:
            running_jobs = database.search(
                Query().queued.exists() & (~Query().finished.exists())
            )
        running_jobs = [r["gdir"] for r in running_jobs]

        return running_jobs

    def _get_finished_jobs(self):
        """"""
        with TinyDB(
            self.directory / f"_{self.scheduler.name}_jobs.json", indent=2
        ) as database:
            finished_jobs = database.search(
                Query().queued.exists() & (Query().finished.exists())
            )
        finished_jobs = [r["gdir"] for r in finished_jobs]

        return finished_jobs

    def _get_retrieved_jobs(self):
        """"""
        with TinyDB(
            self.directory / f"_{self.scheduler.name}_jobs.json", indent=2
        ) as database:
            retrieved_jobs = database.search(
                Query().queued.exists()
                & (Query().finished.exists())
                & Query().retrieved.exists()
            )
        retrieved_jobs = [r["gdir"] for r in retrieved_jobs]

        return retrieved_jobs

    def _get_unretrieved_jobs(self):
        """"""
        with TinyDB(
            self.directory / f"_{self.scheduler.name}_jobs.json", indent=2
        ) as database:
            unretrieved_jobs = database.search(
                (Query().queued.exists() & Query().finished.exists())
                & (~Query().retrieved.exists())
            )
        unretrieved_jobs = [r["gdir"] for r in unretrieved_jobs]

        return unretrieved_jobs

    def get_number_of_running_jobs(self):
        """"""
        running_jobs = self._get_running_jobs()

        return len(running_jobs)

    def as_dict(self) -> dict:
        """"""
        worker_params = {}
        worker_params["potter"] = self.potter.as_dict()
        worker_params["driver"] = self.driver.as_dict()
        worker_params["scheduler"] = self.scheduler.as_dict()

        worker_params = copy.deepcopy(worker_params)

        return worker_params


if __name__ == "__main__":
    ...



#!/usr/bin/env python3
# -*- coding: utf-8 -*

import itertools
import pathlib
import time
import tempfile
from typing import Union, List
import uuid
import warnings
import yaml

from tinydb import Query, TinyDB

from joblib import Parallel, delayed

from ase import Atoms
from ase.io import read, write

from .. import config
from ..potential.manager import AbstractPotentialManager
from .worker import AbstractWorker
from .drive import DriverBasedWorker
from ..data.array import AtomsNDArray
from ..utils.command import CustomTimer
from ..reactor.reactor import AbstractReactor
from .utils import copy_minimal_frames, get_file_md5, read_cache_info


class ReactorBasedWorker(AbstractWorker):

    wdir_prefix: str = "pair" # TODO: cand?

    """Perform the computation of several reactions.
    """

    def __init__(self, potter_, driver_: AbstractReactor=None, scheduler_=None, directory_=None, *args, **kwargs):
        """"""
        self.batchsize = kwargs.pop("batchsize", 1)

        assert isinstance(potter_, AbstractPotentialManager), ""

        self.potter = potter_
        self.driver = driver_
        self.scheduler = scheduler_
        if directory_:
            self.directory = directory_
        
        self.n_jobs = config.NJOBS

        return
    
    def _preprocess(self, structures: Union[List[Atoms], AtomsNDArray]):
        """"""
        # TODO: For now, this only support double-ended methods, 
        #       which means the number of input structures should be even.
        if isinstance(structures, list):
            nstructures = len(structures)
            assert nstructures%2 == 0, "The number of structures should be even."
            pairs = list(
                zip(
                    [structures[i] for i in range(0, nstructures, 2)], 
                    [structures[i] for i in range(1, nstructures, 2)]
                )
            )
        elif isinstance(structures, AtomsNDArray):
            if structures.ndim == 3: # from extract
                assert structures.shape[0] == 2, "Structures must have a shape of (2, ?, ?)."
                pairs = []
                for p in structures:
                    p = [[a for a in s if a is not None][-1] for s in p]
                    pairs.append(p)
                pairs = list(zip(pairs[0], pairs[1]))
            elif structures.ndim == 2: # from extract
                pairs = list(zip(structures[0], structures[1]))
                #raise RuntimeError()
            else:
                pairs = []
                raise RuntimeError()
        else:
            ...

        # - check difference
        processed_dpath = self.directory/"_data"
        processed_dpath.mkdir(exist_ok=True)

        curr_frames, curr_info = copy_minimal_frames(itertools.chain(*pairs))

        # NOTE: handle atoms.info as some codes need energies for IS and FS...
        energies = []
        for i, a in enumerate(itertools.chain(*pairs)):
            try:
                ene = a.get_potential_energy()
                curr_frames[i].info["energy"] = ene
            except:
                ...

        with tempfile.NamedTemporaryFile(mode="w", suffix=".xyz") as tmp:
            write(tmp.name, curr_frames, columns=["symbols", "positions", "move_mask"])

            with open(tmp.name, "rb") as fopen:
                curr_md5 = get_file_md5(fopen)

        _info_data = read_cache_info(self.directory, self.UUIDLEN)
        
        cache_fname = f"{curr_md5}.xyz"
        if (processed_dpath/cache_fname).exists():
            self._print(f"Found file with md5 {curr_md5}")
            self._info_data = _info_data
            start_confid = 0
            for x in self._info_data:
                if x[1] == curr_md5:
                    break
                start_confid += 1
        else:
            write(
                processed_dpath/cache_fname, curr_frames, 
                # columns=["symbols", "positions", "momenta", "tags", "move_mask"]
            )
            # - save current atoms.info and append curr_info to _info_data
            start_confid = len(_info_data)
            content = "{:<12s}  {:<32s}  {:<12s}  {:<12s}  {:<s}\n".format("#id", "MD5", "confid", "step", "wdir")
            for i, (confid, step, wdir) in enumerate(curr_info):
                line = "{:<12d}  {:<32s}  {:<12d}  {:<12d}  {:<s}\n".format(i+start_confid, curr_md5, confid, step, wdir)
                content += line
                _info_data.append(line.strip().split())
            self._info_data = _info_data
            with open(processed_dpath/f"{curr_md5}_info.txt", "w") as fopen:
                fopen.write(content)

        return curr_md5, pairs, start_confid

    def _prepare_batches(self, pairs: List[List[Atoms]], start_confid: int):
        """"""
        nreactions = len(pairs)

        wdirs = [f"{self.wdir_prefix}{i}" for i in range(nreactions)]
        
        # - split reactions into different batches
        starts, ends = self._split_groups(nreactions)

        # - 
        batches = []
        for i, (s, e) in enumerate(zip(starts,ends)):
            curr_indices = range(s, e)
            curr_wdirs = [wdirs[x] for x in curr_indices]
            batches.append([curr_indices, curr_wdirs])

        return batches
    
    def run(self, structures: List[List[Atoms]], *args, **kwargs) -> None:
        """"""
        super().run(*args, **kwargs)

        # - check if the same input structures are provided
        identifier, pairs, start_pairid = self._preprocess(structures)
        batches = self._prepare_batches(pairs, start_confid=start_pairid)

        # - read metadata
        with TinyDB(
            self.directory/f"_{self.scheduler.name}_jobs.json", indent=2
        ) as database:
            queued_jobs = database.search(Query().queued.exists())
        queued_names = [q["gdir"][self.UUIDLEN+1:] for q in queued_jobs]
        queued_input = [q["md5"] for q in queued_jobs]

        # -
        for i, (curr_indices, curr_wdirs) in enumerate(batches):
            # -- set job name
            batch_name = f"group-{i}"
            uid = str(uuid.uuid1())
            job_name = uid + "-" + batch_name

            # -- whether store job info
            if self.scheduler.name != "local":
                if batch_name in queued_names and identifier in queued_input:
                    self._print(f"{batch_name} at {self.directory.name} was submitted.")
                    continue
            else:
                # NOTE: If use local scheduler, always run it again if re-submit
                ...
                
            # -- specify which group this worker is responsible for
            #   if not, then skip
            #   Skip batch here assures the skipped batches will not recorded and 
            #   thus will not affect their execution if several batches run at the same time.
            target_number = kwargs.get("batch", None)
            if isinstance(target_number, int):
                if i != target_number:
                    with CustomTimer(name="run-driver", func=self._print):
                        self._print(
                            f"{time.asctime( time.localtime(time.time()) )} {self.driver.directory.name} batch {i} is skipped..."
                        )
                        continue
                else:
                    ...
            else:
                ...
            
            # -- run batch
            self._irun(
                batch_name, uid, identifier, pairs, curr_indices, curr_wdirs,
                *args, **kwargs
            )

            # - save this batch job to the database
            if identifier not in queued_input:
                with TinyDB(
                    self.directory/f"_{self.scheduler.name}_jobs.json", indent=2
                ) as database:
                    _ = database.insert(
                        dict(
                            uid = uid,
                            md5 = identifier,
                            gdir=job_name, 
                            group_number=i, 
                            wdir_names=curr_wdirs, 
                            queued=True
                        )
                    )


        return
    
    def _irun(
            self, name: str, uid: str, identifier: str, structures, 
            curr_indices, curr_wdirs, *args, **kwargs
        ) -> None:
        """"""
        batch_number = int(name.split("-")[-1])
        if self.scheduler.name == "local":
            with CustomTimer(name="run-reactor", func=self._print):
                # - here the driver is the reactor
                for i, wdir in zip(curr_indices, curr_wdirs):
                    self._print(
                        f"{time.asctime( time.localtime(time.time()) )} {str(wdir)} {self.driver.directory.name} is running..."
                    )
                    self.driver.directory = self.directory/wdir
                    self.driver.reset()
                    _ = self.driver.run(structures[i], read_cache=True)
        else:
            # - save worker file
            worker_params = {}
            worker_params["type"] = "reactor"
            worker_params["driver"] = self.driver.as_dict()
            worker_params["potter"] = self.potter.as_dict()
            worker_params["batchsize"] = self.batchsize

            with open(self.directory/f"worker-{uid}.yaml", "w") as fopen:
                yaml.dump(worker_params, fopen)

            # - save structures
            dataset_path = str((self.directory/"_data"/f"{identifier}.xyz").resolve())

            # - save scheduler file
            jobscript_fname = f"run-{uid}.script"
            self.scheduler.job_name = uid + "-" + name
            self.scheduler.script = self.directory/jobscript_fname

            self.scheduler.user_commands = "gdp -p {} compute {} --batch {}\n".format(
                (self.directory/f"worker-{uid}.yaml").name, 
                #(self.directory/structure_fname).name
                dataset_path, batch_number
            )

            # - TODO: check whether params for scheduler is changed
            self.scheduler.write()
            if self._submit:
                self._print(f"{self.directory.name} JOBID: {self.scheduler.submit()}")
            else:
                self._print(f"{self.directory.name} waits to submit.")
                ...

        return

    def inspect(self, resubmit=False, *args, **kwargs):
        """Check if any job were finished correctly not due to time limit.

        Args:
            resubmit: Check whether submit unfinished jobs.

        """
        self._initialise(*args, **kwargs)
        self._print(f"~~~{self.__class__.__name__}+inspect")

        running_jobs = self._get_running_jobs()

        with TinyDB(
            self.directory/f"_{self.scheduler.name}_jobs.json", indent=2
        ) as database:
            for job_name in running_jobs:
                group_directory = self.directory
                doc_data = database.get(Query().gdir == job_name)
                uid = doc_data["uid"]
                identifier = doc_data["md5"]
                batch = doc_data["group_number"]

                #self.scheduler.set(**{"job-name": job_name})
                self.scheduler.job_name = job_name
                self.scheduler.script = group_directory/f"run-{uid}.script" 

                # -- check whether the jobs if running
                if self.scheduler.is_finished(): # if it is still in the queue
                    # -- valid if the task finished correctly not due to time-limit
                    is_finished = False
                    wdir_names = doc_data["wdir_names"]
                    for x in wdir_names:
                        if not (group_directory/x).exists():
                            # even not start
                            break
                        else:
                            # not converged
                            self.driver.directory = group_directory/x
                            if not self.driver.read_convergence():
                                break
                    else:
                        is_finished = True
                    if is_finished:
                        # -- finished correctly
                        self._print(f"{job_name} is finished...")
                        doc_data = database.get(Query().gdir == job_name)
                        database.update({"finished": True}, doc_ids=[doc_data.doc_id])
                    else:
                        # NOTE: no need to remove unfinished structures
                        #       since the driver would check it
                        if resubmit:
                            if self.scheduler.name != "local":
                                jobid = self.scheduler.submit()
                                self._print(f"{job_name} is re-submitted with JOBID {jobid}.")
                            else:
                                #warnings.warn("Local scheduler does not support re-submit.", UserWarning)
                                frames = read(self.directory/"_data"/f"{identifier}.xyz", ":")
                                self.run(frames, batch=batch)
                else:
                    self._print(f"{job_name} is running...")

        return
    
    def retrieve(self, include_retrieved: bool=False, given_wdirs: List[str]=None, *args, **kwargs):
        """Read results from wdirs.

        Args:
            include_retrieved: Whether include wdirs that are already retrieved.
                              Otherwise, all finished jobs are included.

        """
        self.inspect(*args, **kwargs)
        self._print(f"~~~{self.__class__.__name__}+retrieve")

        # NOTE: sometimes retrieve is used without run
        #self._info_data = self._read_cached_info() # update _info_data

        # - check status and get latest results
        unretrieved_wdirs_ = []
        if not include_retrieved:
            unretrieved_jobs = self._get_unretrieved_jobs()
        else:
            unretrieved_jobs = self._get_finished_jobs()
            
        unretrieved_identifiers = []

        with TinyDB(
            self.directory/f"_{self.scheduler.name}_jobs.json", indent=2
        ) as database:
            for job_name in unretrieved_jobs:
                doc_data = database.get(Query().gdir == job_name)
                unretrieved_identifiers.append(doc_data["md5"])
                unretrieved_wdirs_.extend(
                    self.directory/w for w in doc_data["wdir_names"]
                )
        
        # - get given wdirs
        unretrieved_wdirs = []
        if given_wdirs is not None:
            for wdir in unretrieved_wdirs_:
                wdir_name = wdir.name
                if wdir_name in given_wdirs:
                    unretrieved_wdirs.append(wdir)
        else:
            unretrieved_wdirs = unretrieved_wdirs_

        # - read results
        if unretrieved_wdirs:
            unretrieved_wdirs = [pathlib.Path(x) for x in unretrieved_wdirs]
            results = self._read_results(unretrieved_wdirs, *args, **kwargs)
        else:
            results = []

        with TinyDB(
            self.directory/f"_{self.scheduler.name}_jobs.json", indent=2
        ) as database:
            for job_name in unretrieved_jobs:
                doc_data = database.get(Query().gdir == job_name)
                database.update({"retrieved": True}, doc_ids=[doc_data.doc_id])

        return results

    def _read_results(
        self, unretrieved_wdirs: List[pathlib.Path], *args, **kwargs
    ) -> Union[List[Atoms],List[List[Atoms]]]:
        """Read results from calculation directories.

        Args:
            unretrieved_wdirs: Calculation directories.

        """
        with CustomTimer(name="read-results", func=self._print):
            # NOTE: works for vasp, ...
            results_ = Parallel(n_jobs=self.n_jobs)(
                delayed(self._iread_results)(
                    self.driver, wdir, info_data = None
                ) 
                for wdir in unretrieved_wdirs
            )

            # NOTE: Failed Calcution, One fail, traj fails
            # TODO: check failed...
            #results = []
            #for i, traj_frames in enumerate(results_):
            #    # - sift error structures
            #    if traj_frames:
            #        error_info = traj_frames[0].info.get("error", None)
            #        if error_info:
            #            self._print(f"Found failed calculation at {error_info}...")
            #        else:
            #            results.append(traj_frames)
            #    else:
            #        self._print(f"Found empty calculation at {str(self.directory)} with cand{i}...")

            results = results_

            if results:
                self._print(
                    f"new_trajectories: {len(results)} nframes of the first: {len(results[0])}"
                )
            
        return results
    
    @staticmethod
    def _iread_results(
        driver, wdir, info_data: dict = None
    ) -> List[Atoms]:
        """Extract results from a single directory.

        This must be a staticmethod as it may be pickled by joblib for parallel 
        running.

        Args:
            wdir: Working directory.

        """
        driver.directory = wdir
        # NOTE: name convention, cand1112_field1112_field1112
        confid_ = int(wdir.name.strip("pair").split("_")[0]) # internal name
        if info_data is not None: 
            cache_confid = int(info_data[confid_][2])
            if cache_confid >= 0:
                confid = cache_confid
            else:
                confid = confid_
        else:
            confid = confid_

        # NOTE: always return the entire trajectories
        traj_frames = driver.read_trajectory()
        #for a in traj_frames: # TODO: add to metadata?
        #    a.info["confid"] = confid
        #    a.info["wdir"] = str(wdir.name)
        
        return traj_frames
    
    def as_dict(self) -> dict:
        """"""
        worker_params = super().as_dict()
        worker_params["batchsize"] = self.batchsize

        return worker_params


if __name__ == "__main__":
    ...



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import os
import sys
import logging
import argparse
import pathlib

import numpy as np

# global settings
from . import config
from .core.register import registers, import_all_modules_for_register
from .utils.command import parse_input_file, dict2str


def main():
    # - register
    import_all_modules_for_register()

    description = "gdpx: Generating Deep Potential with Python\n"

    # - arguments 
    parser = argparse.ArgumentParser(
        prog="gdp", 
        description=description
    )

    parser.add_argument(
        "-rs", "--random_seed", default=None, type=int,
        help="global random seed"
    )

    parser.add_argument(
        "-d", "--directory", default=pathlib.Path.cwd(),
        help="working directory"
    )
    
    # the workflow tracker
    parser.add_argument(
        "-p", "--potential", default=None,
        help = "target potential related configuration (json/yaml)"
    )

    parser.add_argument(
        "-nj", "--n_jobs", default = 1, type=int,
        help = "number of processors"
    )

    parser.add_argument(
        "--debug", action="store_true",
        help = "debug mode that gives more information"
    )
    
    parser.add_argument(
        "--log", default="gdp.out",
        help = "logging output file"
    )
    
    # subcommands in the entire workflow 
    subparsers = parser.add_subparsers(
        title="available subcommands", 
        dest="subcommand", 
        help="sub-command help"
    )

    # - run session
    parser_session = subparsers.add_parser(
        "session", help="run gdpy session", 
        description=str(registers.variable)+"\n"+str(registers.operation), 
        formatter_class=argparse.RawDescriptionHelpFormatter
    )
    parser_session.add_argument(
        "SESSION", help="session configuration file (json/yaml)"
    )
    parser_session.add_argument(
        "--feed", default=None, nargs="+", 
        help="session placeholders"
    )
    parser_session.add_argument(
        "--timewait", default=-1, type=float,
        help="waiting time between repeated running"
    )
    
    # - build structures
    parser_build = subparsers.add_parser(
        "build", help="build structures",
        description=str(registers.builder), 
        formatter_class=argparse.RawDescriptionHelpFormatter
    )
    parser_build.add_argument(
        "CONFIG", help="builder configuration file (json/yaml)"
    )
    parser_build.add_argument(
        "-s", "--substrates", default=None,
        help="file that stores substrates (e.g. *.xyz)"
    )
    parser_build.add_argument(
        "-n", "--number", default=1, type=int,
        help="number of structures to build"
    )

    # - convert dataset format
    parser_convert = subparsers.add_parser(
        "convert", help="convert dataset formats",
    )
    parser_convert.add_argument(
        "INPUT", help="path of the input dataset"
    )
    
    # - automatic training
    parser_train = subparsers.add_parser(
        "train", help="automatic training utilities",
        description=str(registers.trainer)+"\n"+str(registers.dataloader), 
        formatter_class=argparse.RawDescriptionHelpFormatter
    )
    parser_train.add_argument(
        "CONFIG", help="training configuration file (json/yaml)"
    )

    # --- compute interface
    parser_compute = subparsers.add_parser(
        "compute", help="compute structures with basic methods (MD, MIN, and ...)",
        description=str(registers.manager),
        formatter_class=argparse.RawDescriptionHelpFormatter
    )
    parser_compute.add_argument(
        "STRUCTURE", nargs="*",
        help="a structure file that stores one or more structures"
    )
    parser_compute.add_argument(
        "-b", "--batch", default=None, type=int,
        help="run selected batch number (useful when queue run)"
    )
    parser_compute.add_argument(
        "--spawn", action="store_true",
        help="If the computation is spawned, it will not save results until all jobs are finished."
    )
    parser_compute.add_argument(
        "--archive", action="store_true",
        help="whether archive computation folders when retrieve"
    )

    # --- expedition interface
    parser_explore = subparsers.add_parser(
        "explore", help="explore structures with advanced methods (GA, MC, and ...)",
        description=str(registers.expedition),
        formatter_class=argparse.RawDescriptionHelpFormatter
    )
    parser_explore.add_argument(
        "CONFIG",
        help="json/yaml file that stores parameters for a task"
    )
    parser_explore.add_argument(
        "--wait", default=None, type=float,
        help="wait time after each run"
    )

    # selection
    parser_select = subparsers.add_parser(
        "select",
        help="apply various selection operations",
        description=str(registers.selector), 
        formatter_class=argparse.RawDescriptionHelpFormatter
    )
    parser_select.add_argument(
        "CONFIG", help="selection configuration file"
    )
    parser_select.add_argument(
        "-s", "--structure", required=True, 
        help="structure generator"
    )

    # --- validation
    parser_validation = subparsers.add_parser(
        "valid", help="validate properties with trained models",
        description=str(registers.validator), 
        formatter_class=argparse.RawDescriptionHelpFormatter
    )
    parser_validation.add_argument(
        "CONFIG", help="validation configuration file"
    )
    
    # === execute 
    args = parser.parse_args()
    
    # - update global configuration
    if args.debug:
        config.logger.setLevel(logging.DEBUG)

    curr_wdir = pathlib.Path(args.directory)
    if not curr_wdir.exists():
        curr_wdir.mkdir(parents=True)

    if args.log:
        logfpath = curr_wdir/args.log
        if logfpath.exists():
            fh = logging.FileHandler(logfpath, mode="a")
        else:
            fh = logging.FileHandler(logfpath, mode="w")
        fh.setFormatter(config.formatter)
        config.logger.addHandler(fh)

    # -- set LOGO
    for line in config.LOGO_LINES:
        config._print(line)

    # -- set njobs
    config.NJOBS = args.n_jobs
    if config.NJOBS != 1:
        config._print(f"Use {config.NJOBS} processors.")
    
    # -- set rng
    # TODO: load random state from a file???
    random_seed = args.random_seed
    if random_seed is None:
        # NOTE: np.random should only be called here once...
        random_seed = np.random.randint(0, 1e8)
    else:
        ...

    config.GRNG = np.random.Generator(np.random.PCG64(random_seed))

    config._print(f"GLOBAL RANDOM SEED : {random_seed}")
    rng_state = config.GRNG.bit_generator.state
    for l in dict2str(rng_state).split("\n"):
        config._print(l)

    # - potential
    potter = None
    if args.potential:
        # a worker or a List of worker
        from .cli.compute import convert_config_to_potter
        potter = convert_config_to_potter(args.potential) 

    # - use subcommands
    if args.subcommand == "session":
        from gdpx.core.session import run_session
        run_session(args.SESSION, args.feed, args.timewait, args.directory)
    elif args.subcommand == "convert":
        from gdpx.data import convert_dataset
        convert_dataset(args.INPUT)
    elif args.subcommand == "train":
        from gdpx.trainer import run_newtrainer
        run_newtrainer(args.CONFIG, args.directory)
    elif args.subcommand == "build":
        build_config = parse_input_file(args.CONFIG)
        from .builder.interface import build_structures
        build_structures(build_config, args.substrates, args.number, args.directory)
    elif args.subcommand == "select":
        from gdpx.selector.interface import run_selection
        run_selection(args.CONFIG, args.structure, args.directory)
    elif args.subcommand == "compute":
        from .cli.compute import run_worker
        run_worker(
            args.STRUCTURE, potter, batch=args.batch, 
            spawn=args.spawn, archive=args.archive, directory=args.directory
        )
    elif args.subcommand == "explore":
        from .expedition.interface import run_expedition
        params = parse_input_file(args.CONFIG)
        run_expedition(params, args.wait, args.directory, potter[0])
    elif args.subcommand == "valid":
        from gdpx.validator import run_validation
        params = parse_input_file(args.CONFIG)
        run_validation(params, args.directory, potter[0])
    else:
        ...
    
    # - report the end random state
    config._print(f"GLOBAL RANDOM SEED : {random_seed}")
    rng_state = config.GRNG.bit_generator.state
    for l in dict2str(rng_state).split("\n"):
        config._print(l)

    return


if __name__ == "__main__":
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


from ase.formula import Formula


def get_atomic_number_list(system_name, type_list):
    """"""
    name_parts = system_name.split("-")
    if len(name_parts) == 1:
        composition = name_parts[0]
    elif len(name_parts) == 2:
        desc, composition = name_parts
    elif len(name_parts) == 3:
        desc, composition, substrate = name_parts
    else:
        raise ValueError(
            f"System name must be as xxx, xxx-xxx, or xxx-xxx-xxx instead of `{system_name}`."
        )

    # print(composition)
    formula = Formula(composition)
    count = formula.count()

    return [count.get(s, 0) for s in type_list][::-1]


def is_a_valid_system_name(system_name: str) -> bool:
    """"""
    is_valid = True

    name_parts = system_name.split("-")
    if len(name_parts) == 1:
        composition = name_parts[0]
    elif len(name_parts) == 2:
        desc, composition = name_parts
    elif len(name_parts) == 3:
        desc, composition, substrate = name_parts
    else:
        composition = ""

    try:
        formula = Formula(composition)
    except:
        is_valid = False

    return is_valid


if __name__ == "__main__":
    ...



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from gdpx.utils.ClusterAndCUR import frames2features
import argparse
from pathlib import Path
from collections import Counter
from itertools import groupby
import pathlib

from typing import Union

import time

import numpy as np

from ase.io import read, write

from tqdm import tqdm
from joblib import Parallel, delayed

import matplotlib as mpl
mpl.use("Agg") #silent mode
from matplotlib import pyplot as plt
try:
    plt.style.use("presentation")
except Exception as e:
    #print("Used default matplotlib style.")
    ...

from ase.io import read, write

from gdpx.utils.data.dpsets import find_systems_set

"""
split total dataset into systemwise one
report data distribution

two ways

from xyz

from dp set
"""


def scatter_hist(x, y, ax, ax_histx, ax_histy):
    # no labels
    ax_histx.tick_params(axis="x", labelbottom=False)
    ax_histy.tick_params(axis="y", labelleft=False)

    # the scatter plot:
    ax.scatter(x, y)

    # now determine nice limits by hand:
    binwidth = 0.25
    xymax = max(np.max(np.abs(x)), np.max(np.abs(y)))
    lim = (int(xymax/binwidth) + 1) * binwidth

    bins = np.arange(-lim, lim + binwidth, binwidth)
    ax_histx.hist(x, bins=bins)
    ax_histy.hist(y, bins=bins, orientation='horizontal')

    return

def plot_stat():
    # definitions for the axes
    left, width = 0.1, 0.65
    bottom, height = 0.1, 0.65
    spacing = 0.005

    rect_scatter = [left, bottom, width, height]
    rect_histx = [left, bottom + height + spacing, width, 0.2]
    rect_histy = [left + width + spacing, bottom, 0.2, height]

    # start with a square Figure
    fig = plt.figure(figsize=(8, 8))

    ax = fig.add_axes(rect_scatter)
    ax_histx = fig.add_axes(rect_histx, sharex=ax)
    ax_histy = fig.add_axes(rect_histy, sharey=ax)

    scatter_hist(energies, energies, ax, ax_histx, ax_histy)

    plt.savefig('miaow.png')

def plot_overview(
    frames, 
    ftol: float =0.05
):
    """ force tolerance 0.05 for perfectly optimised structures 
        while 0.10 for nearly optimised ones...
    """
    force_list = []
    optimised_frames = []
    for atoms in frames:
        forces = atoms.get_forces()
        maxforce = np.max(forces)
        if maxforce < ftol:
            optimised_frames.append(atoms)
        force_list.extend(forces.flatten().tolist())

    energies = [atoms.get_potential_energy() for atoms in frames]
    print(len(energies))
    opt_energies = [atoms.get_potential_energy() for atoms in optimised_frames]
    print(len(opt_energies))

    return energies, force_list


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument(
        '-f', '--file',
        default='./test.xyz', help='xyz file'
    )
    parser.add_argument(
        '-c', '--chosen', nargs='*', required=True,
        help='chosen system, O 1 Pt 36'
    )
    parser.add_argument(
        '-nj', '--njobs', type=int,
        default=4, help='upper limit on number of directories'
    )
    parser.add_argument(
        '-ns', '--NotSave', action="store_false",
        help='save xyz for dpsets'
    )
    parser.add_argument(
        '-na', '--nall', action="store_false",
        help='not save train and test togather'
    )

    args = parser.parse_args()

    # format chosen
    if len(args.chosen) % 2 == 0:
        chosen = ''.join(args.chosen)
        print('Choose system %s' %chosen)
    else:
        raise ValueError('Wrong System Format')

    #find_systems(args.file, args.njobs)
    raw_path = Path(args.file)
    for p in raw_path.iterdir():
        if p.name == chosen:
            cur_system = p
            break
    
    print('find ', cur_system)
    if args.NotSave:
        train_xyz = pathlib.Path(cur_system.stem + '-train.xyz')
        test_xyz = pathlib.Path(cur_system.stem + '-test.xyz')
        all_xyz = pathlib.Path(cur_system.stem + '-all.xyz')
        if train_xyz.exists() and test_xyz.exists():
            print('read existed frames...')
            train_frames = read(train_xyz, ':')
            test_frames = read(test_xyz, ':')
        else:
            print('read and save frames...')
            train_frames, test_frames = find_systems_set(cur_system)
            write(train_xyz, train_frames)
            write(test_xyz, test_frames)
            all_frames = []
            all_frames.extend(train_frames)
            all_frames.extend(test_frames)
            if not args.nall:
                write(all_xyz, all_frames)
    else:
        train_frames, test_frames = find_systems_set(cur_system)

    # prepare to plot histogram
    train_energies, train_forces = plot_overview(train_frames, ftol=0.10)
    test_energies, test_forces = plot_overview(test_frames, ftol=0.10)

    fig, axarr = plt.subplots(nrows=1, ncols=2, figsize=(16,12))
    axarr = axarr.flatten()
    
    num_bins = 50

    ax = axarr[0]
    n, bins, patches = ax.hist(
        train_energies, num_bins, density=False,
        label = 'Trainset'
    )
    n, bins, patches = ax.hist(
        test_energies, num_bins, density=False,
        label = 'Testset'
    )
    ax.legend()

    ax = axarr[1]
    n, bins, patches = ax.hist(
        train_forces, num_bins, density=False,
        label = 'Trainset'
    )
    n, bins, patches = ax.hist(
        test_forces, num_bins, density=False,
        label = 'Testset'
    )
    ax.legend()


    plt.savefig('ana.png')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from ase.ga.particle_crossovers import CutSpliceCrossover
from ase.ga.cutandsplicepairing import CutAndSplicePairing

from gdpx.core.register import registers

registers.builder.register("cut_and_splice")(CutAndSplicePairing)
registers.builder.register("cut_and_splice_cluster")(CutSpliceCrossover)



if __name__ == "__main__":
    ...


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import abc
import logging
import pathlib
from typing import NoReturn, Union, Callable

from gdpx import config


class AbstractValidator(abc.ABC):

    _print: Callable = config._print
    _debug: Callable = config._debug

    _directory = pathlib.Path.cwd()

    def __init__(self, directory: Union[str,pathlib.Path]="./", *args, **kwargs):
        """
        """
        self.directory = directory

        self.njobs = config.NJOBS

        return
    
    @property
    def directory(self):
        """"""

        return self._directory

    @directory.setter
    def directory(self, directory_):
        """"""
        directory_ = pathlib.Path(directory_)
        self._directory = directory_

        return

    @abc.abstractmethod
    def run(self, *args, **kwargs):
        """"""
        if not self.directory.exists():
            self.directory.mkdir(parents=True)

        return


if __name__ == "__main__":
    ...